from contextlib import contextmanager
from functools import wraps, update_wrapper

from .exceptions import TException


class LockstepSyncMismatch(TException):
    template = "Expected {lockstep_name} to be {expected_step}, but it is {actual_step}"


class _LockstepInvocation(object):
    def __init__(self, name, generator):
        self._name = name
        self._generator = generator
        self._current_step = 'just-started'

    def step_next(self, step):
        """Progress to the next step, confirming it's the specified one"""

        try:
            self._current_step, value = self._next_step_and_value()
        except StopIteration:
            raise LockstepSyncMismatch(lockstep_name=self._name,
                                       expected_step=step,
                                       actual_step='finished')
        if self._current_step != step:
            raise LockstepSyncMismatch(lockstep_name=self._name,
                                       expected_step=step,
                                       actual_step=self._current_step)
        return value

    def step_until(self, step):
        """Progress until we get to the specified step, confirming that it exist"""

        while self._current_step != step:
            try:
                self._current_step, value = self._next_step_and_value()
            except StopIteration:
                raise LockstepSyncMismatch(lockstep_name=self._name,
                                           expected_step=step,
                                           actual_step='finished')
        return value

    def step_all(self):
        """Progress through all the remaining steps"""

        for self._current_step in self._generator:
            pass

    def __iter__(self):
        yield from self._generator
        self._current_step = 'finished'

    def _next_step_and_value(self):
        yield_result = next(self._generator)
        if isinstance(yield_result, tuple) and len(yield_result) == 2:
            return yield_result
        else:
            return yield_result, None

    def __str__(self):
        return '%s<%s>' % (self._name, self._current_step)


class lockstep(object):
    """
    Synchronize a coroutine that runs a process step-by-step.

    Decorate a generator that yields step names to create a context manager.

    * Use like a regular method(that returns `None`).
    * Use ``.lockstep(...)`` on the method to get a context manager. The context
      object has a ``.step_next``/``.step_until`` methods that must be called, in
      order, with all expected step names, to make the generator progress to
      each step.
    * Yield from the context object to embed the lockstep inside a bigger
      lockstep function.

    Example::

        @lockstep
        def my_process():
            # things before step A
            yield 'A'
            # things between step A and step B
            yield 'B'
            # things between step B and step C
            yield 'C'
            # things between step C and step D
            yield 'D'
            # things after step D

        my_process()  # just run it like a normal function, ignoring the lockstep

        with my_process.lockstep() as process:
            process.step_next('A')  # go to next step - A
            process.step_until('C')  # go through steps until you reach C
            process.step_all()  # go through all remaining steps until the end

        @lockstep
        def bigger_process():
            yield 'X'

            # Embed `my_process`'s steps inside `bigger_process`:
            with my_process.lockstep() as process:
                yield from process

            yield 'Y'

    :note: When used class methods, put the ``@classmethod`` decorator **below**
           the ``@lockstep`` decorator::

            class Foo:
                @lockstep
                @classmethod
                def process(cls):
                    # ...
    """

    def __init__(self, generator_func):
        self.generator_func = generator_func
        update_wrapper(self, generator_func)

    @contextmanager
    def lockstep(self, *args, **kwargs):
        generator = self.generator_func(*args, **kwargs)
        invocation = _LockstepInvocation(self.generator_func.__name__, generator)

        yield invocation

        try:
            step_not_taken, _ = invocation._next_step_and_value()
        except StopIteration:
            # all is well - all steps were exhausted
            invocation._current_step = 'finished'
        else:
            raise LockstepSyncMismatch(lockstep_name=invocation._name,
                                       expected_step='finished',
                                       actual_step=step_not_taken)

    def __call__(self, *args, **kwargs):
        with self.lockstep(*args, **kwargs) as process:
            process.step_all()

    def __get__(self, instance, owner=None):
        func = self.generator_func.__get__(instance, owner)
        if func is self.generator_func:
            return self
        else:
            return lockstep(func)



try:
    import gevent
    from gevent.monkey import is_module_patched
except ImportError:
    def is_module_patched(*_, **__):
        return False


from logging import getLogger
import threading   # this will be reloaded after patching

import time
import sys


from easypy.humanize import format_thread_stack
from ._multithreading_init import _set_thread_uuid, _set_main_uuid


_logger = getLogger('gevent')


main_thread_ident_before_patching = threading.main_thread().ident

HUB = None

HOGGING_TIMEOUT = 1


def apply_patch(hogging_detection=False, real_threads=1):
    _logger.info('applying gevent patch (%s real threads)', real_threads)

    # real_threads is 1 by default so it will be possible to run watch_threads concurrently
    if hogging_detection:
        real_threads += 1

    _patch_module_locks()

    import gevent
    import gevent.monkey

    for m in ["easypy.threadtree", "easypy.concurrency"]:
        assert m not in sys.modules, "Must apply the gevent patch before importing %s" % m

    gevent.monkey.patch_all(Event=True, sys=True)

    _unpatch_logging_handlers_lock()

    global HUB
    HUB = gevent.get_hub()

    global threading
    import threading
    for thread in threading.enumerate():
        _set_thread_uuid(thread.ident)
    _set_main_uuid()  # the patched threading has a new ident for the main thread

    # this will declutter the thread dumps from gevent/greenlet frames
    from .threadtree import _BOOTSTRAPPERS
    import gevent, gevent.threading, gevent.greenlet
    _BOOTSTRAPPERS.update([gevent, gevent.threading, gevent.greenlet])

    if hogging_detection:
        import greenlet
        greenlet.settrace(lambda *args: _greenlet_trace_func(*args))
        defer_to_thread(detect_hogging, 'detect-hogging')


def _patch_module_locks():
    # gevent will not patch existing locks (including ModuleLocks) when it's not single threaded
    # our solution is to monkey patch the release method for ModuleLocks objects
    # we assume that patching is done early enough so no other locks are present

    import importlib
    _old_release = importlib._bootstrap._ModuleLock.release

    def _release(*args, **kw):
        lock = args[0]
        if lock.owner == main_thread_ident_before_patching:
            lock.owner = threading.main_thread().ident
        _old_release(*args, **kw)

    importlib._bootstrap._ModuleLock.release = _release


def _unpatch_logging_handlers_lock():
    # we dont want to use logger locks since those are used by both real thread and gevent greenlets
    # switching from one to the other will cause gevent hub to throw an exception

    RLock = gevent.monkey.saved['threading']['_CRLock']

    def create_unpatched_lock_for_handler(handler):
        handler.lock = RLock()

    import logging
    # patch future handlers
    logging.Handler.createLock = create_unpatched_lock_for_handler
    for handler in logging._handlers.values():
        if handler.lock:
            handler.lock = RLock()

    try:
        import logbook.handlers
    except ImportError:
        pass
    else:
        # patch future handlers
        logbook.handlers.new_fine_grained_lock = RLock
        for handler in logbook.handlers.Handler.stack_manager.iter_context_objects():
            handler.lock = RLock()


def _greenlet_trace_func(event, args):
    pass


def detect_hogging():
    did_switch = True

    current_running_greenlet = HUB

    def mark_switch(event, args):
        nonlocal did_switch
        nonlocal current_running_greenlet
        if event != 'switch':
            return
        did_switch = True
        current_running_greenlet = args[1]  # args = [origin_greenlet , target_greenlet

    global _greenlet_trace_func
    _greenlet_trace_func = mark_switch

    current_blocker_time = 0
    last_warning_time = 0

    while True:
        non_gevent_sleep(HOGGING_TIMEOUT)
        if did_switch:
            # all good
            pass
        elif current_running_greenlet == HUB:
            # it's ok for the hub to block if all greenlet wait on async io
            pass
        else:
            current_blocker_time += HOGGING_TIMEOUT
            if current_blocker_time < last_warning_time * 2:
                continue  # dont dump too much warnings - decay exponentialy until exploding after FAIL_BLOCK_TIME_SEC
            for thread in threading.enumerate():
                if getattr(thread, '_greenlet', None) == current_running_greenlet:
                    _logger.info('RED<<greenlet hogger detected (%s seconds):>>', current_blocker_time)
                    _logger.debug('thread stuck: %s', thread)
                    break
            else:
                _logger.info('RED<<unknown greenlet hogger detected (%s seconds):>>', current_blocker_time)
                _logger.debug('greenlet stuck (no corresponding thread found): %s', current_running_greenlet)
                _logger.debug('hub is: %s', HUB)
            func = _logger.debug if current_blocker_time < 5 * HOGGING_TIMEOUT else _logger.info
            func("Stack:\n%s", format_thread_stack(sys._current_frames()[main_thread_ident_before_patching]))
            last_warning_time = current_blocker_time
            continue

        current_blocker_time = 0
        last_warning_time = 0
        did_switch = False


def non_gevent_sleep(timeout):
    try:
        gevent.monkey.saved['time']['sleep'](timeout)
    except KeyError:
        time.sleep(timeout)


def defer_to_thread(func, threadname):

    def run():
        _set_thread_uuid(threading.get_ident(), parent_uuid)
        _logger.debug('Starting job in real thread: %s', threadname or "<anonymous>")
        gevent.spawn(func)  # running via gevent ensures we have a Hub
        gevent.wait()
        _logger.debug('ready for the next job')

    from .threadtree import get_thread_uuid
    parent_uuid = get_thread_uuid(threading.current_thread())
    pool = gevent.get_hub().threadpool
    return pool.spawn(run).wait



import pytest
from easypy.collections import separate
from easypy.collections import ListCollection, SimpleObjectCollection, partial_dict, UNIQUE, ObjectNotFound, TooManyObjectsFound
from easypy.bunch import Bunch
from collections import Counter


class Obj(Bunch):
    def __repr__(self):
        return "%(name)s:%(id)s:%(v)s" % self


L = ListCollection(Obj(name=n, id=i, v=v) for n, i, v in zip("aabcdddeff", "1234567890", "xxxyyyxxyz") for _ in range(100))


def test_collection_filter():
    lst = ListCollection("abcdef")
    assert lst.filtered(lambda c: c == 'a').sample(1) == ['a']


def test_collection_reprs():

    def check(lst):
        str(lst)
        repr(lst)

    check(L)
    check(L.filtered(id=5))

    NL = ListCollection(L, name='lst')

    check(NL)
    check(NL.filtered(id=5))

    O = SimpleObjectCollection(L, ID_ATTRIBUTE='name')

    check(O)
    check(O.filtered(id=5))

    NO = SimpleObjectCollection(L, ID_ATTRIBUTE='name', name='objs')

    check(NO)
    check(NO.filtered(id=5))


def test_partial_dict():
    assert partial_dict({'a': 1, 'b': 2, 'c': 3}, ['a', 'b']) == {'a': 1, 'b': 2}


def test_separate1():
    a, b = separate(range(5), key=lambda n: n < 3)
    assert a == [0, 1, 2]
    assert b == [3, 4]


def test_separate2():
    a, b = separate(range(5))
    assert a == [1, 2, 3, 4]
    assert b == [0]


def test_collection_sample():
    l = ListCollection("abcdef")
    assert len(l.sample(2.0)) == 2

    with pytest.raises(AssertionError):
        l.sample(1.5)


def test_collection_select():
    assert len(L.select(name='a', id='1')) == 100


def test_collection_select_no_unique():
    with pytest.raises(AssertionError):
        L.select(name=UNIQUE)


def test_collection_sample_too_much():
    len(L.select(name='a', id='2').sample(100)) == 100
    with pytest.raises(ObjectNotFound):
        L.select(name='a', id='2').sample(101)


def test_collection_sample_too_many():
    len(L.select(name='a', id='2').sample(100)) == 100
    with pytest.raises(TooManyObjectsFound):
        L.get(name='a', id='2')


def test_collection_sample_unique0():
    assert not L.sample(0, name=UNIQUE)


def test_collection_sample_unique1():
    s = L.sample(3, name=UNIQUE)
    assert len({b.name for b in s}) == 3


def test_collection_sample_unique2():
    x, = L.sample(1, name=UNIQUE, id='1')
    assert x.id == '1'


def test_collection_sample_unique3():
    s = L.sample(6, name=UNIQUE, id=UNIQUE)
    assert len({(b.name, b.id) for b in s}) == 6


def test_collection_sample_unique4():
    with pytest.raises(ObjectNotFound):
        L.sample(7, name=UNIQUE)  # too many


def test_collection_sample_unique5():
    s = L.sample(3, name=UNIQUE, id=UNIQUE, v=UNIQUE)
    assert len({(b.name, b.id) for b in s}) == 3


def test_collection_sample_unique_diverse():
    x = Counter(repr(x) for _ in range(100) for x in L.sample(1, name=UNIQUE))
    assert len(x) == 10


def test_collections_slicing():
    L = ListCollection("abcdef")
    assert L[0] == 'a'
    assert L[-1] == 'f'
    assert L[:2] == list('ab')
    assert L[-2:] == list('ef')
    assert L[::2] == list('ace')
    assert L[::-2] == list('fdb')


def test_filters_order():
    l = ListCollection([Bunch(a='b', b=5), Bunch(a='c', c='c')])
    filterd_l = l.filtered(a='b').select(lambda o: o.b > 4)
    assert len(filterd_l) == 1

    with pytest.raises(AttributeError):
        filterd_l = l.select(lambda o: o.b > 4)


def test_simple_object_collection():
    S = SimpleObjectCollection(L, ID_ATTRIBUTE='id')
    assert S.get_next(S['5']) == S['6']
    assert S.get_prev(S['5']) == S['4']



"""
This module helps you with handling and generating collections of items.
"""


from __future__ import absolute_import
import collections
from numbers import Integral
from itertools import chain, islice
from functools import partial
import inspect
import random
from .predicates import make_predicate
from .tokens import UNIQUE
from .decorations import parametrizeable_decorator
from .exceptions import TException

import sys

from collections import OrderedDict as PythonOrderedDict
SUPPORT_GET_SIBLING = True

if sys.version_info[:2] >= (3, 5):
    # In order to support 'get_prev' and 'get_next', we need access to OrderedDict's internal .__map,
    # which we don't have in the C-implementation of the class in Python3.5
    # This hack allows us to get to the pythonic implemenation of OrderedDict
    try:
        from test.support import import_fresh_module
        PythonOrderedDict = import_fresh_module('collections', blocked=['_collections']).OrderedDict
    except ImportError:
        # this happens on 3.6.7-8 due to a bug (https://bugzilla.redhat.com/show_bug.cgi?id=1651245)
        SUPPORT_GET_SIBLING = False


def _format_predicate(pred):
    args = inspect.formatargspec(*inspect.getargspec(pred))
    return "<lambda>" if pred.__name__ == "<lambda>" else "{pred.__name__}{args}".format(**locals())


def _format_filter_string(preds, filters):
    returned = []
    returned.extend(map(_format_predicate, preds))
    returned.extend("{!s}={!r}".format(*item) for item in filters.items())
    return ", ".join(returned)


class ObjectNotFound(TException, LookupError):
    template = "Object not found in {collection!r}, filtered by {_filters}"

    def __init__(self, collection, preds=(), filters={}, **kw):
        preds = getattr(collection, "preds", ()) + preds
        filters = dict(getattr(collection, "filters", {}), **filters)
        filter_str = _format_filter_string(preds, filters)
        super().__init__(collection=collection, _filters=filter_str, **kw)


class TooManyObjectsFound(ObjectNotFound):
    template = "Too many objects ({_found}) found in {collection!r}, {_filters}"


class NotEnoughObjects(ObjectNotFound):
    template = "Not enough objects found in {collection!r}, {_filters}"


class Repr(str):
    """
    A string that represents itself without quotes
    """
    __repr__ = str.__str__


class defaultlist(list):
    """
    Like defaultdict, but for a list.
    Any access to an index that isn't populated will fill it with the value returned by the default factory:

    >>> d = defaultlist(int)
    >>> d[3] += 1
    >>> d
    [0, 0, 0, 1]
    """

    def __init__(self, default_factory):
        self.default_factory = default_factory
        super().__init__()

    def _fill(self, index):
        while len(self) <= index or (len(self) == 0 and index == -1):
            self.append(self.default_factory())

    def __setitem__(self, index, value):
        self._fill(index)
        list.__setitem__(self, index, value)

    def __getitem__(self, index):
        self._fill(index)
        return list.__getitem__(self, index)


def _get_value(obj, attr):
    value = getattr(obj, attr)
    if callable(value):
        value = value()
    return value


def _validate(obj, key, value):
    assert value is not UNIQUE, "Can't specify %s except with '.sample()'" % UNIQUE
    actual = _get_value(obj, key)
    return actual == value


def filters_to_predicates(filters):
    """
    Converts the given dict to a list of predicate callables.
    Used by the various *Collection classes in this module.

    >>> from .bunch import Bunch
    >>> [pred] = filters_to_predicates(dict(foo='bar'))

    >>> obj1 = Bunch(foo='bar')
    >>> pred(obj1)  # checks!
    True

    >>> obj2 = Bunch(bar='foo', foo='bad')
    >>> pred(obj2)  # checks!
    False
    """
    return [make_predicate(partial(_validate, key=key, value=value))
            for key, value in filters.items()]


def filtered(objects, preds, filters):
    """
    Yields the object that pass the predicates (callables), and filters (dict).
    Used by the various *Collection classes in this module.

    >>> from .bunch import Bunch
    >>> l = [Bunch(a=a, b=b) for a in range(3) for b in range(3)]
    >>> f = list(filtered(l, [lambda o: o.a==o.b], dict(a=1)))
    >>> f
    [Bunch(a=1, b=1)]
    """
    preds = [make_predicate(p) for p in preds]
    preds += filters_to_predicates(filters)

    # don't collapse into one line so easier to debug
    for obj in objects:
        first_exception = None
        for pred in preds:
            try:
                if not pred(obj):
                    break
            except Exception as exc:
                if first_exception is None:
                    first_exception = exc
        else:
            if first_exception is not None:
                raise first_exception
            yield obj


def uniquify(objects, num, attrs):
    """
    Find ``num`` objects (in random order) that are unique in their `attrs` attributes.
    Returns empty list if no objects are found to satisfy the uniqueness condition.

    >>> from .bunch import Bunch

    >>> l = [
    ...     Bunch(first='Adam', last='Brody', age='n/a'),
    ...     Bunch(first='Adam', last='Levine', age='n/a'),
    ...     Bunch(first='Adam', last='Sandler', age='n/a'),
    ...     Bunch(first='Amy', last='Adams', age='n/a'),
    ...     Bunch(first='Amy', last='Poehler', age='n/a'),
    ...     Bunch(first='Amy', last='Winehouse', age='n/a')]

    >>> o1, o2 = uniquify(l, 2, ["first", "last"])
    >>> (o1.first != o2.first) and (o1.last != o2.last)
    True

    >>> list(uniquify(l, 3, ["first", "last"]))
    []

    >>> list(uniquify(l, 2, ["first", "age"]))
    []
    """

    object_map = {}
    partitions = collections.defaultdict(lambda: collections.defaultdict(set))

    for object in objects:
        k = id(object)
        object_map[k] = object
        for attr in attrs:
            value = _get_value(object, attr)
            partitions[attr][value].add(k)

    if not all(len(group.values()) >= num for group in partitions.values()):
        return []

    def gen(available, collected=[]):
        if len(collected) == num:
            yield [object_map[k] for k in collected]
            return

        for k in shuffled(available):
            excluded = {i for attr in attrs for i in partitions[attr][_get_value(object_map[k], attr)]}
            remain = available - {k} - excluded
            yield from gen(remain, collected=collected + [k])

    try:
        return next(gen(set(object_map.keys())))
    except StopIteration:
        return []


# Inspired by code written by Rotem Yaari

class ObjectCollectionBase(object):

    def __init__(self, name=None):
        super().__init__()
        self.name = name

    def __iter__(self):
        raise NotImplementedError()

    def __len__(self):
        returned = 0
        for _ in self:
            returned += 1
        return returned

    def __nonzero__(self):
        for _ in self:
            return True
        return False

    def _new(self, items):
        return self.__class__(items)

    def _getitem(self, index):
        if isinstance(index, Integral):
            if index < 0:
                return list(self)[index]
            try:
                return next(islice(self, index, index+1))
            except StopIteration:
                raise LookupError(index)
        elif isinstance(index, slice):
            if any(i < 0 for i in (index.start, index.stop, index.step) if i is not None):
                # can't use islice when indices are negative
                items = list(self)
                return self._new(items[slice(index.start, index.stop, index.step)])
            try:
                return self._new(islice(self, index.start, index.stop, index.step))
            except StopIteration:
                raise LookupError(index)
        else:
            return self.get(index)

    def __getitem__(self, index):
        return self._getitem(index)

    def __repr__(self):
        if not self.name:
            items = all_items = list(self)
            if len(items) > 3:
                items = items[:2] + [Repr("...")] + items[-1:]
            return "{0.__class__.__name__}({items}, size={size})".format(self, items=items, size=len(all_items))
        return "<{0.__class__.__name__} '{0.name}'>".format(self)

    def __str__(self):
        items = ", ".join(map(str, self))
        if not self.name:
            return "<{}>".format(items)
        return "<'{}': {}>".format(self.name, items)

    def _repr_pretty_(self, p, cycle):
        from easypy.colors import DARK_CYAN
        # used by IPython
        class_name = self.__class__.__name__
        if cycle:
            p.text('%s(...)' % DARK_CYAN(class_name))
            return
        prefix = '%s(' % DARK_CYAN(class_name)
        with p.group(len(class_name) + 1, prefix, ')'):
            for idx, v in enumerate(self):
                if idx:
                    p.text(',')
                    p.breakable()
                p.pretty(v)

    def get_by_key(self, key):
        return self.get(key=key)

    def get_by_uid(self, uid):
        return self.get_by_key(uid)

    def safe_get_by_key(self, key):
        try:
            return self.get_by_key(key)
        except ObjectNotFound:
            return None

    def safe_get_by_uid(self, uid):
        return self.safe_get_by_key(uid)

    def get_by_keys(self, keys):
        gen = map(self.get_by_key, keys)
        try:
            return type(keys)(gen)  # return something of the same type as keys
        except TypeError:  # keys is generator type - return a generator
            return gen

    def iter_filtered(self, *preds, **filters):
        _shuffle = filters.pop("_shuffle", False)
        if _shuffle:
            objects = list(self)
            random.shuffle(objects)
        else:
            objects = self
        return filtered(objects, preds, filters)

    def select(self, *preds, **filters):
        return ListCollection(self.iter_filtered(*preds, **filters))

    def safe_get(self, *preds, **filters):
        # this python fu avoids filtering the entire list, while still checking for one item
        matching = [obj for obj, _ in zip(self.iter_filtered(*preds, **filters), range(2))]
        if len(matching) > 1:
            raise TooManyObjectsFound(self, preds, filters, _found=len(matching))
        return matching[0] if matching else None

    def get(self, *preds, **filters):
        # this python fu avoids filtering the entire list, while still checking for one item
        matching = [obj for obj, _ in zip(self.iter_filtered(*preds, **filters), range(2))]
        if len(matching) == 0:
            raise ObjectNotFound(self, preds, filters)
        if len(matching) != 1:
            raise TooManyObjectsFound(self, preds, filters, _found=len(matching))
        return matching[0]

    def choose(self, *preds, **filters):
        for obj in self.iter_filtered(_shuffle=True, *preds, **filters):
            return obj
        raise ObjectNotFound(self, preds, filters)

    def safe_choose(self, *preds, **filters):
        try:
            return self.choose(*preds, **filters)
        except ObjectNotFound:
            return None

    def menu(self, multiselect=False):
        """Open termenu to interactively select item(s) from the collection"""
        from termenu import show_menu
        return show_menu(self.name, list(self), multiselect=multiselect)

    def filtered(self, *preds, **filters):
        return FilterCollection(self, preds, filters)

    def shuffled(self):
        return self.sample(len(self))

    def sorted(self, key=None):
        return self._new(sorted(self, key=key))

    def sample(self, num, *preds, **filters):
        """
        If 'num' is negative, leaves 'num' out of the sample:

            >>> collection = ListCollection([1, 2, 3])
            >>> sample = collection.sample(-1)
            >>> len(sample)
            2
            >>> all(i in collection for i in sample)
            True
        """
        if num < 0:
            num += len(self)

        if isinstance(num, float):
            num, d = divmod(num, 1)
            assert d == 0, "Can't sample a non-integer number of items"
            num = int(num)

        if not num:
            return self._new([])

        uniquifiers = []
        for k, v in list(filters.items()):
            if v is UNIQUE:
                filters.pop(k)
                uniquifiers.append(k)

        if uniquifiers:
            if (preds or filters):
                matching = self.iter_filtered(_shuffle=True, *preds, **filters)
            else:
                matching = self

            matching = uniquify(matching, num, uniquifiers)

        else:
            # use 'zip' as a limiter
            matching = [obj for _, obj in zip(range(num), self.iter_filtered(_shuffle=True, *preds, **filters))]

        if len(matching) < num:
            raise NotEnoughObjects(self, preds, filters, needed=num)

        return self._new(matching)

    def _choose_sampling_size(self, minimum=0, maximum=None):
        if minimum < 0:
            minimum += len(self)
        if maximum is None:
            maximum = len(self)
        elif maximum < 0:
            maximum += len(self)
        maximum = min(maximum, len(self))
        return random.randint(minimum, maximum)

    def sample_some(self, minimum=0, maximum=None):
        sample_size = self._choose_sampling_size(minimum, maximum)
        return self.sample(sample_size)

    def without(self, *items_to_exclude):
        def filter(item):
            return item not in items_to_exclude
        return self.filtered(filter)

    @property
    def L(self):
        return list(self)

    @property
    def M(self):
        from .concurrency import MultiObject
        return MultiObject(self)

    def __add__(self, collection):
        return AggregateCollection([self, collection])


class AggregateCollection(ObjectCollectionBase):
    """
    Dynamically aggregate other collections into one chained collection
    """

    def __init__(self, collections, name=None):
        super().__init__()
        self._collections = collections
        self.name = name or "+".join((c.name or 'unnamed') for c in collections)

    def __iter__(self):
        return chain(*self._collections)

    def __add__(self, collection):
        return AggregateCollection(self._collections + [collection])

    def _new(self, items):
        return ListCollection(items)


class ListCollection(list, ObjectCollectionBase):
    """
    A simple list-based implementation of the ObjectCollection protocol
    """
    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop('name', None)
        super().__init__(*args, **kwargs)

    def pop_some(self, minimum=0, maximum=None):
        sample_size = self._choose_sampling_size(minimum, maximum)
        return self._new([self.pop() for i in range(sample_size)])

    def __getitem__(self, index):
        return self._getitem(index)

    def __repr__(self):
        if not self.name:
            return ObjectCollectionBase.__repr__(self)
        return "<'{0.name}', size={size}>".format(self, size=len(self))


class SimpleObjectCollection(ObjectCollectionBase):
    """
    An ObjectCollection with fast lookup based on an ID attribute found on the objects.
    """

    ID_ATTRIBUTE = 'uid'

    class Collectable():
        """
        Objects can inherit this mixin class and will then get a ``.collection`` attribute
        that points back to the collection they were added to
        """
        collection = None

    def __init__(self, objs=(), ID_ATTRIBUTE=None, backref=False, name=None):
        super().__init__(name=name)
        if ID_ATTRIBUTE:
            self.ID_ATTRIBUTE = ID_ATTRIBUTE
        self._objects = PythonOrderedDict()
        for obj in objs:
            self.add(obj, backref=backref)

    def __repr__(self):
        if not self.name:
            return ObjectCollectionBase.__repr__(self)
        return "<'{0.name}', size={size}>".format(self, size=len(self))

    def _new(self, items):
        return self.__class__(items, self.ID_ATTRIBUTE)

    def _get_object_uid(self, obj):
        return getattr(obj, self.ID_ATTRIBUTE)

    def _add_and_get_uid(self, obj):
        uid = self._get_object_uid(obj)
        self._objects[uid] = obj
        return uid, obj

    def add(self, obj, backref=False):
        if backref:
            assert isinstance(obj, self.Collectable), "Can add backref only to Collectable classes"
            assert obj.collection is None or obj.collection is self, \
                "Collectable object not allowed to be added to multiple collections"
            obj.collection = self
        return self._add_and_get_uid(obj)[1]

    def _remove_and_get_uid(self, obj):
        uid = self._get_object_uid(obj)
        return uid, self._objects.pop(uid)

    def remove(self, obj):
        return self._remove_and_get_uid(obj)[1]

    def clear(self):
        self._objects.clear()

    def remove_by_uid(self, uid):
        obj = self._objects.pop(uid)
        return obj

    def get(self, *preds, **filters):
        if len(preds) == 1 and not filters and not callable(preds[0]):
            key, = preds
            return self.get_by_key(key)
        return super().get(*preds, **filters)

    def get_by_key(self, key):
        if key not in self._objects:
            raise ObjectNotFound(self, (), dict(key=key),)
        return self._objects[key]

    def index(self, obj):
        lookup_uid = self._get_object_uid(obj)
        if lookup_uid not in self._objects:
            raise ValueError("%r not in %r" % (obj, self))
        for i, uid in enumerate(self._objects.keys()):
            if lookup_uid == uid:
                return i

    if SUPPORT_GET_SIBLING:
        def get_next(self, obj):
            uid = self._get_object_uid(obj)
            if uid not in self._objects:
                raise ObjectNotFound(self, (), dict(key=uid),)
            next_uid = self._objects._OrderedDict__map[uid].next
            try:
                key = next_uid.key
            except AttributeError:
                key = next_uid.next.key
            return self._objects[key]

        def get_prev(self, obj):
            uid = self._get_object_uid(obj)
            if uid not in self._objects:
                raise ObjectNotFound(self, (), dict(key=uid),)
            prev_uid = self._objects._OrderedDict__map[uid].prev
            try:
                key = prev_uid.key
            except AttributeError:
                key = prev_uid.prev.key
            return self._objects[key]
    else:
        def get_next(self, obj):
            from warning import warn
            warn("Couldn't import a pure-python OrderedDict so this will be a O(n) operation")

            from itertools import cycle
            l1, l2 = cycle(self._objects.values()), cycle(self._objects.values())
            next(l2)
            for a, b in zip(l1, l2):
                if obj == a:
                    return b

        def get_prev(self, obj):
            from warning import warn
            warn("Couldn't import a pure-python OrderedDict so this will be a O(n) operation")

            from itertools import cycle
            l1, l2 = cycle(self._objects.values()), cycle(self._objects.values())
            next(l2)
            for a, b in zip(l1, l2):
                if obj == b:
                    return a

    def keys(self):
        return self._objects.keys()

    def __iter__(self):
        return iter(self._objects.values())

    def __len__(self):
        return len(self._objects)

    @property
    def M(self):
        from .concurrency import MultiObject
        return MultiObject(self, list(self._objects.keys()))


class FilterCollection(ObjectCollectionBase):
    """
    A dynamic collection that filters another ``base`` collection according to the
    given predicates and filters.
    """

    def __init__(self, base, preds, filters, parent=None, name=None):
        super().__init__(name=name)
        self.base = base
        self.parent = parent if parent is not None else base
        self.preds = preds
        self.filters = filters

    def _new(self, items):
        if hasattr(self.base, 'ID_ATTRIBUTE'):
            return self.base.__class__(items, ID_ATTRIBUTE=self.base.ID_ATTRIBUTE)
        return self.base.__class__(items)

    def __repr__(self):
        if self.name and self.base.name:
            return "<'{self.name}', based on '{self.base.name}'>".format(**locals())
        elif self.name:
            return "<'{self.name}', based on a {self.base.__class__.__name__}>".format(**locals())
        elif not self.name:
            filters_str = _format_filter_string(self.preds, self.filters)
            if self.base.name:
                return "<Filtered collection based on '{self.base.name}' ({filters_str})>".format(**locals())
            else:
                return "<Filtered collection based on an anonymous '{self.base.__class__.__name__}' ({filters_str})>".format(**locals())
        else:
            return ObjectCollectionBase.__repr__(self)

    def __iter__(self):
        return self.iter_filtered()

    def iter_filtered(self, *preds, **filters):
        _shuffle = filters.pop("_shuffle", False)
        combined_preds = list(self.preds)
        combined_preds += preds
        for k, v in self.filters.items():
            if k in filters:
                assert v == filters[k], "Two different values provided for the same filter key!"
        filters.update(self.filters)
        return self.parent.iter_filtered(_shuffle=_shuffle, *combined_preds, **filters)

    def filtered(self, *preds, **filters):
        return FilterCollection(self.base, preds, filters, parent=self)

    def get(self, *preds, **filters):
        if len(preds) == 1 and not filters and not callable(preds[0]):
            key, = preds
            return self.get_by_key(key)
        else:
            return super().get(*preds, **filters)

    def get_by_key(self, key):
        try:
            obj = self.base.get_by_key(key)
        except ObjectNotFound:
            raise ObjectNotFound(self, key=key) from None
        for obj in filtered([obj], self.preds, self.filters):
            return obj
        raise ObjectNotFound(self, key=key)

    def get_next(self, obj):
        while True:
            next_obj = self.base.get_next(obj)
            if any(filtered([next_obj], self.preds, self.filters)):
                return next_obj
            obj = next_obj

    def __getitem__(self, uid):
        if not isinstance(uid, str):
            return super().__getitem__(uid)

        try:
            obj = self.base[uid]
        except ObjectNotFound:
            raise ObjectNotFound(self, key=uid) from None

        for obj in filtered([obj], self.preds, self.filters):
            return obj

        raise ObjectNotFound(self, key=uid)


def TypeFilterCollection(base, type):
    """
    A collection that filters another collection according to the specified type (class)
    """
    return base.filtered(lambda obj: isinstance(obj, type))


class IteratorBasedCollection(ObjectCollectionBase):
    """
    An iterator-based collection. Each time the collection is read, the iterator function is called again,
    to provide a fresh iterator object.
    """

    def __init__(self, iterator_func):
        super().__init__()
        self._iterator_func = iterator_func

    def __iter__(self):
        return self._iterator_func()


class IndexedObjectCollection(SimpleObjectCollection):
    """
    Work-in-progress

    An indexed collection, allowing fast lookup of object by using multiple key indices.
    """

    def __init__(self, objs=(), keys=(), **kwargs):
        self._indices = {key: collections.defaultdict(set) for key in keys}
        super().__init__(objs, **kwargs)

    def _new(self, items):
        return self.__class__(items, keys=self._indices.keys(), ID_ATTRIBUTE=self.ID_ATTRIBUTE)

    def add(self, obj):
        uid, obj = self._add_and_get_uid(obj)
        for key, uid_map in self._indices.items():
            value = getattr(obj, key)
            uid_map[value].add(uid)

    def remove(self, obj):
        uid, obj = self._remove_and_get_uid(obj)
        for uid_map in self._indices.values():
            for uids in uid_map.values():
                uids.discard(uid)
        return obj

    def clear(self):
        self._objects.clear()
        for idx in self._indices.values():
            idx.clear()

    def remove_by_uid(self, uid):
        return self.remove(self._objects[uid])

    def get(self, *preds, **filters):
        if len(preds) == 1 and not filters and not callable(preds[0]):
            key, = preds
            return self.get_by_key(key)
        return super().get(*preds, **filters)

    def iter_filtered(self, *preds, **filters):
        _shuffle = filters.pop("_shuffle", False)
        indexed = set(filters) & set(self._indices)
        unindexed = set(filters) - indexed
        if indexed:
            filtered_keys = set.intersection(*(self._indices.get(key, {}).get(filters[key], set()) for key in indexed))
            objects = (self.get_by_key(key) for key in filtered_keys)
        else:
            objects = self
        if _shuffle:
            objects = list(objects)
            random.shuffle(objects)
        if not (preds or unindexed):
            return objects
        filters = {k: filters[k] for k in unindexed}
        return filtered(objects, preds, filters)


def grouped(sequence, key=None, transform=None):
    """
    Parse the sequence into groups, according to key:

        >>> ret = grouped(range(10), lambda n: n % 3)
        >>> ret[0]
        [0, 3, 6, 9]
        >>> ret[1]
        [1, 4, 7]
        >>> ret[2]
        [2, 5, 8]
    """
    groups = {}
    if not key:
        key = lambda x: x
    if not transform:
        transform = lambda x: x
    for item in sequence:
        groups.setdefault(key(item), []).append(transform(item))
    return groups


def separate(sequence, key=None):
    """
    Partition the sequence into items that match and items that don't:

        >>> above_three, three_or_less = separate(range(10), lambda n: n > 3)

        >>> print(above_three)
        [4, 5, 6, 7, 8, 9]

        >>> print(three_or_less)
        [0, 1, 2, 3]
    """
    if not key:
        key = lambda x: x
    groups = grouped(sequence, key=lambda e: bool(key(e)))
    return groups.get(True, []), groups.get(False, [])


def iterable(obj):
    return isinstance(obj, collections.Iterable) and not isinstance(obj, (str, bytes, dict))


def ilistify(obj):
    """
    Normalize input into an iterable:

        >>> list(ilistify(1))
        [1]
        >>> list(ilistify([1, 2, 3]))
        [1, 2, 3]
    """
    if not iterable(obj):
        yield obj
    else:
        yield from obj


def listify(obj):
    """
    Normalize input into a list:
        >>> listify(1)
        [1]
        >>> listify([1, 2])
        [1, 2]
    """
    return list(ilistify(obj))


def chunkify(sequence, size):
    """
    Chunk a sequence into equal-size chunks (except for the last chunk):

        >>> ["".join(p) for p in chunkify('abcdefghijklmnopqrstuvwxyz', 7)]
        ['abcdefg', 'hijklmn', 'opqrstu', 'vwxyz']
    """
    sequence = iter(sequence)
    while True:
        chunk = [e for _, e in zip(range(size), sequence)]
        if not chunk:
            return
        yield chunk


def partial_dict(d, keys):
    """
    Returns a new dict with a subset of the original items.

    >>> d = partial_dict({'a': 1, 'b': 2, 'c': 3}, ['a', 'b'])
    >>> d == {'a': 1, 'b': 2}
    True
    """
    return {k: d[k] for k in keys}


def intersected_dict(d, keys):
    """
    Returns a new dict with a subset of the original items.
    Items that are specified in keys but not in the original dict are discarded

    >>> d = intersected_dict({'a': 1, 'b': 2, 'c': 3}, ['a', 'b', 'z'])
    >>> d == {'a': 1, 'b': 2}
    True
    """
    return {k: d[k] for k in keys if k in d}


def dicts_to_table(dicts, index_header="idx"):
    """
    Convert a sequence of homogenuous dicts into a table (2D list),
    with the dict keys present in the first row:

    >>> dict_list = [
    ...     dict(a=1, b=2, c=3),
    ...     dict(a=7, b=8, c=9)]
    >>> for row in dicts_to_table(dict_list, index_header='R'):
    ...     print(*row, sep="|")
    R|a|b|c
    0|1|2|3
    1|7|8|9

    >>> dicts_dict = dict(
    ...     X=dict(a=1, b=2, c=3),
    ...     Y=dict(a=7, b=8, c=9))
    >>> for row in dicts_to_table(dicts_dict, index_header='R'):
    ...     print(*row, sep="|")
    R|a|b|c
    X|1|2|3
    Y|7|8|9

    """

    if isinstance(dicts, dict):
        _, sample_row = next(iter(dicts.items()))
        rows = ((key, dicts[key]) for key in sorted(dicts))
    else:
        sample_row = dicts[0]
        rows = enumerate(dicts)
    headers = sorted(sample_row.keys())
    table = [[index_header] + headers]
    for key, row in rows:
        table.append([key] + [row[header] for header in headers])
    return table


def shuffled(items):
    """
    Return a shuffled version of the input sequence
    """
    items = list(items)
    random.shuffle(items)
    return items


class SlidingWindow(list):
    """
    A list that maintains a constant size, popping old items as new items are appended (FIFO)
    """
    def __init__(self, *args, size):
        self.size = size
        super().__init__(*args)

    def append(self, item):
        super().append(item)
        if len(self) > self.size:
            self.pop(0)


@parametrizeable_decorator
def as_list(generator, sort_by=None):
    """
    Forces a generator to output a list.

    For when writing a generator is more elegant::

    >>> @as_list(sort_by=lambda n: -n)
    ... def g():
    ...     yield 2
    ...     yield 1
    ...     yield from range(2)
    >>> g()
    [2, 1, 1, 0]

    """

    def inner(*args, **kwargs):
        l = list(generator(*args, **kwargs))
        if sort_by:
            l.sort(key=sort_by)
        return l
    return inner


def takesome(generator, max=None, min=0):
    """
    Take between ``min`` and ``max`` items from the generator.
    Throws a ValueError if the generator didn't yield ``min`` items.

    >>> list(takesome("abcdef", 3))
    ['a', 'b', 'c']

    >>> list(takesome("abcdef", 10))
    ['a', 'b', 'c', 'd', 'e', 'f']

    >>> list(takesome("abcdef", 0))
    []

    >>> list(takesome("abcdef", min=8))
    Traceback (most recent call last):
        ...
    ValueError: Not enough items in sequence (wanted 8, got 6)
    """

    if max is not None and max < min:
        raise ValueError("'max' must be great than 'min'")

    items = [p for p, _ in zip(generator, range(min))]

    if len(items) < min:
        raise ValueError("Not enough items in sequence (wanted {}, got {})".format(min, len(items)))

    yield from items

    if max is None:
        yield from generator
        return

    remain = max - min
    for p, _ in zip(generator, range(remain)):
        yield p


if __name__ == "__main__":
    import doctest
    doctest.testmod()



class Predicate(object):

    def test(self, obj):
        raise NotImplementedError()

    def __call__(self, obj):
        return self.test(obj)

    def __eq__(self, obj):
        return self.test(obj)

    def __ne__(self, obj):
        return not self == obj

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)

    def __str__(self):
        return self.describe()

    def __repr__(self):
        return "<%s>" % self

    def describe(self, variable="X"):
        return self._describe(variable) + "?"

    def _describe(self, variable):
        raise NotImplementedError()


class FunctionPredicate(Predicate):

    def __init__(self, func, description=None):
        super(FunctionPredicate, self).__init__()
        self.func = func
        if description is None:
            description = func.__doc__
        self.description = description

    def test(self, obj):
        if isinstance(obj, FunctionPredicate):
            return obj.func is self.func
        else:
            return self.func(obj)

    def _describe(self, variable):
        if self.description:
            return self.description % dict(var=variable)
        else:
            return "%s(%s)" % (self.func,variable)


class Equality(Predicate):

    def __init__(self, value):
        super(Equality, self).__init__()
        self.value = value

    def test(self, obj):
        if isinstance(obj, Equality):
            return obj.value == self.value
        else:
            return obj == self.value

    def _describe(self, variable):
        return "%s==%s" % (variable, str(self.value))


Inequality = lambda value: Not(Equality(value))


class Or(Predicate):

    def __init__(self, *preds):
        super(Or, self).__init__()
        self.preds = map(make_predicate, preds)

    def test(self, obj):
        return any(p == obj for p in self.preds)

    def _describe(self, variable):
        return " OR ".join("(%s)" % pred._describe(variable) for pred in self.preds)


class And(Predicate):

    def __init__(self, *preds):
        super(And, self).__init__()
        self.preds = map(make_predicate, preds)

    def test(self, obj):
        return all(p == obj for p in self.preds)

    def _describe(self, variable):
        return " AND ".join("(%s)" % pred._describe(variable) for pred in self.preds)


class Not(Predicate):

    def __init__(self, pred):
        super(Not, self).__init__()
        self.pred = make_predicate(pred)

    def test(self, obj):
        return not self.pred == obj

    def _describe(self, variable):
        return "NOT(%s)" % (self.pred._describe(variable),)


class _Dummy(Predicate):

    def __init__(self, retval, description=""):
        self.retval = retval
        self.description = description

    def test(self, other):
        return self.retval

    def _describe(self, variable):
        return self.description

IGNORE = _Dummy(True, "ANYTHING")
FAIL = _Dummy(False, "NOTHING")


def make_predicate(expr):
    """
    common utility for making various expressions into predicates
    """
    if isinstance(expr, Predicate):
        return expr
    elif isinstance(expr, type):
        return FunctionPredicate(lambda obj, type=expr: isinstance(obj, type))
    elif callable(expr):
        return FunctionPredicate(expr)
    else:
        return Equality(expr)

P = make_predicate



"""
This module is about 'property' descriptors.
"""

import sys
from functools import wraps
from easypy.caching import cached_property  # public import, for back-compat


_builtin_property = property


def safe_property(fget=None, fset=None, fdel=None, doc=None):
    """
    A pythonic property which raises a RuntimeError when an attribute error is raised within it.
    This fixes an issue in python where AttributeErrors that occur anywhere _within_ 'property' functions
    are effectively suppressed, and converted to AttributeErrors for the property itself. This is confusing
    for the debugger, and also leads to unintended fallback calls to a __getattr__ if defined

    >>> def i_raise_an_exception():
    ...     raise AttributeError("blap")

    >>> class Test(object):
    ...     def some_prop(self):
    ...         return i_raise_an_exception()
    ...     def __getattr__(self, attr):
    ...         assert False
    ...     prop = property(some_prop)
    ...     safe_prop = safe_property(some_prop)
    >>> t = Test()
    >>> t.prop
    Traceback (most recent call last):
     ...
    AssertionError
    >>> t.safe_prop
    Traceback (most recent call last):
     ...
     AttributeError: blap
     ...
     During handling of the above exception, another exception occurred:
     ...
     Traceback (most recent call last):
     ...
    RuntimeError: Attribute error within a property (blap)
    """
    if fget is not None:
        @wraps(fget)
        def callable(*args, **kwargs):
            try:
                return fget(*args, **kwargs)
            except AttributeError:
                _, exc, tb = sys.exc_info()
                raise RuntimeError("Attribute error within a property (%s)" % exc).with_traceback(tb)
        return _builtin_property(callable, fset, fdel, doc)
    else:
        return _builtin_property(fget, fset, fdel, doc)



import pytest
import random

from easypy.random import random_nice_name, random_filename


def test_random_nice_name():
    for _ in range(20):
        length = random.randint(64, 85)
        entropy = random.randint(1, 3)
        sep = random.choice(['_', '..'])
        name = random_nice_name(max_length=length, entropy=entropy, sep=sep)
        assert len(name) <= length


def test_random_nice_name_raises():
    with pytest.raises(ValueError):
        random_nice_name(max_length=10, entropy=3)


def test_random_filename():
    fn = random_filename(10)
    assert len(fn) == 10

    fn = random_filename((10, 11))
    assert 10 <= len(fn) <= 11

    fn = random_filename((11, 11))
    assert len(fn) == 11



import pytest

from easypy.lockstep import lockstep, LockstepSyncMismatch


def test_lockstep_side_effects():
    calculation_result = 0

    @lockstep
    def simple_calculation(number):
        nonlocal calculation_result

        calculation_result = number
        yield 'SET_NUMBER'

        calculation_result *= 2
        yield 'MULTIPLY_IT_BY_TWO'

        calculation_result += 5
        yield 'ADD_FIVE'

    with simple_calculation.lockstep(5) as calculation:
        calculation.step_next('SET_NUMBER')
        assert calculation_result == 5

        calculation.step_next('MULTIPLY_IT_BY_TWO')
        assert calculation_result == 10

        calculation.step_next('ADD_FIVE')
        assert calculation_result == 15


def test_lockstep_run_as_function():
    calculation_result = 0

    @lockstep
    def simple_calculation(number):
        nonlocal calculation_result

        calculation_result = number
        yield 'SET_NUMBER'

        calculation_result *= 2
        yield 'MULTIPLY_IT_BY_TWO'

        calculation_result += 5
        yield 'ADD_FIVE'

    simple_calculation(10)
    assert calculation_result == 25


def test_lockstep_class_method():
    class SimpleCalculation():
        def __init__(self, number):
            self.calculation_result = number

        @lockstep
        def calculation(self):
            self.calculation_result *= 2
            yield 'MULTIPLY_IT_BY_TWO'

            self.calculation_result += 5
            yield 'ADD_FIVE'

    simple_calculation = SimpleCalculation(5)
    with simple_calculation.calculation.lockstep() as calculation:
        assert simple_calculation.calculation_result == 5

        calculation.step_next('MULTIPLY_IT_BY_TWO')
        assert simple_calculation.calculation_result == 10

        calculation.step_next('ADD_FIVE')
        assert simple_calculation.calculation_result == 15
    assert simple_calculation.calculation_result == 15

    # run as function
    simple_calculation2 = SimpleCalculation(10)
    assert simple_calculation2.calculation_result == 10
    simple_calculation2.calculation()
    assert simple_calculation2.calculation_result == 25


def test_lockstep_wrong_step_name():
    @lockstep
    def process():
        yield 'STEP_1'
        yield 'STEP_2'
        yield 'STEP_3'

    with pytest.raises(LockstepSyncMismatch) as excinfo:
        with process.lockstep() as process:
            process.step_next('STEP_1')
            process.step_next('STEP_TWO')
            process.step_next('STEP_3')

    assert excinfo.value.expected_step == 'STEP_TWO'
    assert excinfo.value.actual_step == 'STEP_2'


def test_lockstep_not_exhausted():
    @lockstep
    def process():
        yield 'STEP_1'
        yield 'STEP_2'
        yield 'STEP_3'

    with pytest.raises(LockstepSyncMismatch) as excinfo:
        with process.lockstep() as process:
            process.step_next('STEP_1')
            process.step_next('STEP_2')

    assert excinfo.value.expected_step == 'finished'
    assert excinfo.value.actual_step == 'STEP_3'


def test_lockstep_exhausted_prematurely():
    @lockstep
    def process():
        yield 'STEP_1'
        yield 'STEP_2'

    with pytest.raises(LockstepSyncMismatch) as excinfo:
        with process.lockstep() as process:
            process.step_next('STEP_1')
            process.step_next('STEP_2')
            process.step_next('STEP_3')

    assert excinfo.value.expected_step == 'STEP_3'
    assert excinfo.value.actual_step == 'finished'


def test_lockstep_exhaust():
    finished = False

    @lockstep
    def process():
        nonlocal finished

        yield 'STEP_1'
        yield 'STEP_2'
        yield 'STEP_3'

        finished = True

    assert not finished
    with process.lockstep() as process:
        assert not finished
        process.step_all()
        assert finished
    assert finished


def test_lockstep_yielded_values():
    @lockstep
    def process():
        yield 'STEP_1', 1
        yield 'STEP_2'
        yield 'STEP_3', 3

    with process.lockstep() as process:
        assert process.step_next('STEP_1') == 1
        assert process.step_next('STEP_2') is None
        assert process.step_next('STEP_3') == 3


def test_lockstep_nested():
    @lockstep
    def internal_process():
        yield 'INTERNAL_1'
        yield 'INTERNAL_2'

    @lockstep
    def external_process():
        yield 'EXTERNAL_1'
        with internal_process.lockstep() as process:
            yield from process
        yield 'EXTERNAL_2'

    with external_process.lockstep() as process:
        process.step_next('EXTERNAL_1')
        process.step_next('INTERNAL_1')
        process.step_next('INTERNAL_2')
        process.step_next('EXTERNAL_2')


def test_lockstep_step_util():
    @lockstep
    def process():
        yield 'STEP_1'
        yield 'STEP_2'
        yield 'STEP_3'

    with process.lockstep() as process:
        process.step_until('STEP_3')


def test_lockstep_step_util_wrong_order():
    @lockstep
    def process():
        yield 'STEP_1'
        yield 'STEP_2'
        yield 'STEP_3'

    with pytest.raises(LockstepSyncMismatch) as excinfo:
        with process.lockstep() as process:
            process.step_until('STEP_2')
            process.step_until('STEP_1')

    assert excinfo.value.expected_step == 'STEP_1'
    assert excinfo.value.actual_step == 'finished'


def test_lockstep_as_static_and_class_methods():
    class Foo:
        @lockstep
        def process1(self, out):
            out.append(1)
            yield 'STEP'
            out.append(2)

        @lockstep
        @classmethod
        def process2(cls, out):
            out.append(1)
            yield 'STEP'
            out.append(2)

        @lockstep
        @staticmethod
        def process3(out):
            out.append(1)
            yield 'STEP'
            out.append(2)

    print()

    def check_method_call(method):
        out = []
        method(out)
        assert out == [1, 2]

    check_method_call(Foo().process1)
    check_method_call(Foo().process2)
    check_method_call(Foo().process3)

    def check_method_lockstep(method):
        method.lockstep
        out = []
        with method.lockstep(out) as process:
            assert out == []
            process.step_until('STEP')
            assert out == [1]
        assert out == [1, 2]

    check_method_lockstep(Foo().process1)
    check_method_lockstep(Foo.process2)
    check_method_lockstep(Foo.process3)



from easypy.humanize import from_hexdump, hexdump, IndentableTextBuffer, format_table


_SAMPLE_DATA = b'J\x9c\xe8Z!\xc2\xe6\x8b\xa0\x01\xcb\xc3.x]\x11\x9bsC\x1c\xb2\xcd\xb3\x9eM\xf7\x13`\xc8\xce\xf8g1H&\xe2\x9b'     \
    b'\xd1\xa8\xfd\x14\x08U\x175\xc7\x03q\xac\xda\xe6)q}}T44\x9e\xb5;\xf1.\xf6*\x16\xba\xe0~m\x96o\xb8\xa4Tl\x96\x8a\xc7'    \
    b'\x9a\xc9\xc4\xf2\xb1\x9e\x13\x0b\xe2i\xc6\xd8\x92\xde\xfabn6\xea\xf5_y>\x15\xc5\xd5\xa0\x05\xbd\xea\xb8\xba\x80+P'     \
    b'\xa7\xd8\xad\xbf\x91<\xca\xc5\x94\xe6\xfc-\xab4ABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABAB'     \
    b'ABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABA'     \
    b'BABABABABABABABABABABABABABABABABABABABABCABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABAB'    \
    b'ABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABABA'     \
    b'BABABABABABABABABABABABABABABABABABABABABAB\xdc)n/\x9aNy\x9f\x03\xc7j\x14\x08\x1a\x08\x91@\xad\xac\xa9(\x1a\x8b\x9f'   \
    b'\x81\xb0us\x87\x9e4\xf9\x991w39\xd0\x98XokH\xa6\xc9Rv\xbc\xac\x90;\xac\x83\xc8\xba`V\xa9\xc3u\xb2\xccV\x9d\x06\xb3'    \
    b'\xf0\x1e\xb4K\x10\x9c\x83\xdc\xe7\xcb\x0c\x9a\x8c\x80\x010\x8ca\xf85Z\x9c'


def test_hexdump_functions():
    assert from_hexdump(hexdump(_SAMPLE_DATA)) == _SAMPLE_DATA
    assert from_hexdump(hexdump(_SAMPLE_DATA, 24, 2)) == _SAMPLE_DATA
    assert from_hexdump(hexdump(_SAMPLE_DATA, 16, 1, False)) == _SAMPLE_DATA
    assert from_hexdump(hexdump(_SAMPLE_DATA, 4, 4)) == _SAMPLE_DATA

    assert _SAMPLE_DATA.decode("hexdump_24_2") == hexdump(_SAMPLE_DATA, 24, 2)
    assert hexdump(_SAMPLE_DATA, 24, 2).encode("hexdump") == _SAMPLE_DATA


def test_indentable_text_buffer():
    from io import StringIO

    buff = IndentableTextBuffer("Exc")
    buff.write("a")
    buff.write("b")
    with buff.indent("Header2"):
        buff.write(hexdump(_SAMPLE_DATA, 24, 8))
    buff.write("hello")
    buff.write("world")
    with buff.indent("Header2"):
        # buff.write(format_in_columns([str(i) for i in range(100)], 50))
        with buff.indent("This should be pruned away"):
            with buff.indent("This should be pruned away"):
                pass
        with buff.indent("Header3"):
            buff.write("text3")
        buff.write("text2")

    f = StringIO()
    buff.render(prune=True, textual=True, width=120, file=f)
    assert open("tests/indentable_buffer1.txt", "r").read() == f.getvalue()

    f = StringIO()
    buff.render(prune=True, textual=False, width=40, overflow="ignore", file=f)
    assert open("tests/indentable_buffer2.txt", "r").read() == f.getvalue()

    f = StringIO()
    buff.render(prune=True, textual=False, width=40, edges=False, file=f)
    assert open("tests/indentable_buffer3.txt", "r").read() == f.getvalue()


def test_format_table_with_titles():
    table = [
        'abc',
        range(3),
        [None, True, False],
        [dict(x='x'), b'bytes', 'string']
    ]

    output = (
        "a         |b       |c     \n"
        "--------------------------\n"
        "         0|       1|     2\n"
        "None      |True    |False \n"
        "{'x': 'x'}|b'bytes'|string\n")

    assert output == format_table(table)


def test_format_table_without_titles():
    table = [
        'abc',
        range(3),
        [None, True, False],
        [dict(x='x'), b'bytes', 'string']
    ]

    output = (
        "a         |b       |c     \n"
        "         0|       1|     2\n"
        "None      |True    |False \n"
        "{'x': 'x'}|b'bytes'|string\n")

    assert output == format_table(table, titles=False)



import os
import sys
import time
import shelve
import inspect
from threading import RLock
from logging import getLogger
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from functools import wraps, _make_key, partial, lru_cache, update_wrapper

from .decorations import parametrizeable_decorator, DecoratingDescriptor
from .collections import ilistify
from .misc import kwargs_resilient
from .deprecation import deprecated
from .units import HOUR
from .tokens import DELETED, NO_DEFAULT
from .humanize import yesno_to_bool

_logger = getLogger(__name__)

try:
    from _gdbm import error as GDBMException
except:  # noqa
    try:
        from _dbm import error as GDBMException
    except:  # noqa
        from dbm import error as GDBMException


DISABLE_CACHING_PERSISTENCE = yesno_to_bool(os.getenv("EASYPY_DISABLE_CACHING_PERSISTENCE", "no"))


class PersistentCache(object):
    """
    A memoizer that stores its cache persistantly using shelve.

    :param path: location of cache shelve file.
    :param version: modify to deprecate old cahed data
    :param expiration: expiration in seconds for the entire cache. ``None`` to disable expiration.

    Example::

        >>> CACHE = PersistentCache("/tmp/cache", version=1, expiration=60)

        >>> @CACHE
        ... def fib(n):
        ...     a, b = 0, 1
        ...     while a < n:
        ...         a, b = b, a+b
        ...     return a
    """

    def __init__(self, path, version=None, expiration=4 * HOUR, ignored_keywords=None):
        self.path = path
        self.version = version
        if version is not None:
            self.path = "{self.path}.v{version}".format(**locals())
        self.expiration = expiration
        self.lock = RLock()
        self.ignored_keywords = set(ilistify(ignored_keywords)) if ignored_keywords else set()

    @contextmanager
    def db_opened(self, lock=False):
        if DISABLE_CACHING_PERSISTENCE:
            yield {}
            return

        from .resilience import retrying
        with ExitStack() as stack:
            with self.lock:
                try:
                    db = stack.enter_context(
                        retrying(3, acceptable=GDBMException, sleep=5)(shelve.open)(self.path))
                except Exception:
                    try:
                        os.unlink(self.path)
                    except FileNotFoundError:
                        pass
                    try:
                        db = stack.enter_context(shelve.open(self.path))
                    except Exception:
                        _logger.warning("Could not open PersistentCache: %s", self.path)
                        db = {}

            if lock:
                stack.enter_context(self.lock)
            yield db

    def set(self, key, value):
        with self.db_opened(lock=True) as db:
            if value == DELETED:
                del db[key]
            else:
                db[key] = (value, time.time())

    def get(self, key, default=NO_DEFAULT):
        try:
            with self.db_opened() as db:
                value, timestamp = db[key]
            if not self.expiration:
                pass
            elif (timestamp + self.expiration) <= time.time():
                raise KeyError()
            return value
        except KeyError:
            if default is NO_DEFAULT:
                raise
            else:
                return default

    def __call__(self, func=None, *, validator=None):
        if validator and not func:
            return partial(self.__call__, validator=validator)

        validator = validator and kwargs_resilient(validator)

        @wraps(func)
        def inner(*args, **kwargs):
            key_kwargs = {k: v for k, v in kwargs.items() if k not in self.ignored_keywords}
            key = str(_make_key(
                (func.__module__, func.__qualname__,) + args, key_kwargs,
                typed=False, kwd_mark=("_KWD_MARK_",)))

            try:
                value = self.get(key)
            except KeyError:
                pass
            else:
                if not validator:
                    return value
                validated_value = validator(value, args=args, kwargs=kwargs)
                if validated_value:
                    self.set(key, validated_value)
                    return validated_value
                self.set(key, DELETED)
                return inner(*args, **kwargs)
            ret = func(*args, **kwargs)
            self.set(key, ret)
            return ret
        inner.clear_cache = self.clear
        return inner

    def clear(self):
        with self.db_opened() as db:
            db.clear()


@deprecated(message="Please use easypy.caching.locking_cache")
def locking_lru_cache(maxsize=128, typed=False):
    """
    DEPRECATED!

    An lru cache decorator with a lock, to prevent concurrent invocations and allow reusing from cache.

    :param maxsize: LRU cache maximum size, defaults to 128
    :type maxsize: number, optional
    :param typed: If typed is set to true, function arguments of different types will be cached separately. defaults to False.
    :type typed: bool, optional
    """

    def deco(func):
        caching_func = lru_cache(maxsize, typed)(func)
        func._main_lock = RLock()
        func._keyed_locks = defaultdict(RLock)

        @wraps(func)
        def inner(*args, **kwargs):
            key = _make_key(args, kwargs, typed=typed)
            with func._main_lock:
                key_lock = func._keyed_locks[key]
            with key_lock:
                return caching_func(*args, **kwargs)

        @wraps(caching_func.cache_clear)
        def clear():
            with func._main_lock:
                return caching_func.cache_clear()

        inner.cache_clear = clear
        return inner

    return deco


if sys.version_info < (3, 5):
    def _apply_defaults(bound_arguments):
        """
        Set default values for missing arguments. (from Python3.5)
        """
        from collections import OrderedDict
        from inspect import _empty, _VAR_POSITIONAL, _VAR_KEYWORD

        arguments = bound_arguments.arguments
        new_arguments = []
        for name, param in bound_arguments._signature.parameters.items():
            try:
                new_arguments.append((name, arguments[name]))
            except KeyError:
                if param.default is not _empty:
                    val = param.default
                elif param.kind is _VAR_POSITIONAL:
                    val = ()
                elif param.kind is _VAR_KEYWORD:
                    val = {}
                else:
                    # This BoundArguments was likely produced by
                    # Signature.bind_partial().
                    continue
                new_arguments.append((name, val))
        bound_arguments.arguments = OrderedDict(new_arguments)
else:
    def _apply_defaults(bound_arguments):
        bound_arguments.apply_defaults()


class _TimeCache(DecoratingDescriptor):
    def __init__(self, func, **kwargs):
        update_wrapper(self, func)  # this needs to be first to avoid overriding attributes we set
        super().__init__(func=func, cached=True)
        self.func = func
        self.kwargs = kwargs
        self.expiration = kwargs['expiration']
        self.typed = kwargs['typed']
        self.get_ts_func = kwargs['get_ts_func']
        self.log_recalculation = kwargs['log_recalculation']
        self.ignored_keywords = kwargs['ignored_keywords']

        if self.ignored_keywords:
            assert not kwargs['key_func'], "can't specify both `ignored_keywords` AND `key_func`"
            self.ignored_keywords = set(ilistify(self.ignored_keywords))

            def key_func(**kw):
                return tuple(v for k, v in sorted(kw.items()) if k not in self.ignored_keywords)
        else:
            key_func = kwargs['key_func']

        self.NOT_FOUND = object()
        self.NOT_CACHED = self.NOT_FOUND, 0

        self.cache = {}
        self.main_lock = RLock()
        self.keyed_locks = defaultdict(RLock)

        if key_func:
            sig = inspect.signature(func)

            def make_key(args, kwargs):
                bound = sig.bind(*args, **kwargs)
                _apply_defaults(bound)
                return kwargs_resilient(key_func)(**bound.arguments)
        else:
            def make_key(args, kwargs):
                return _make_key(args, kwargs, typed=self.typed)

        self.make_key = make_key

    def __call__(self, *args, **kwargs):
        key = self.make_key(args, kwargs)

        with self.main_lock:
            key_lock = self.keyed_locks[key]
        with key_lock:
            result, ts = self.cache.get(key, self.NOT_CACHED)

            if self.expiration <= 0:
                pass  # nothing to fuss with, cache does not expire
            elif result is self.NOT_FOUND:
                pass  # cache is empty
            elif self.get_ts_func() - ts >= self.expiration:
                # cache expired
                result = self.NOT_FOUND
                del self.cache[key]

            if result is self.NOT_FOUND:
                if self.log_recalculation:
                    _logger.debug('time cache expired, calculating new value for %s', self.__name__)
                result = self.func(*args, **kwargs)
                self.cache[key] = result, self.get_ts_func()

            return result

    def cache_clear(self):
        with self.main_lock:
            for key, lock in dict(self.keyed_locks).items():
                with lock:
                    self.cache.pop(key, None)

    def cache_pop(self, *args, **kwargs):
        key = self.make_key(args, kwargs)
        self.keyed_locks.pop(key, None)
        return self.cache.pop(key, None)

    def _decorate(self, method, instance, owner):
        return type(self)(method, **self.kwargs)


def timecache(expiration=0, typed=False, get_ts_func=time.time, log_recalculation=False, ignored_keywords=None, key_func=None):
    """
    A thread-safe cache decorator with time expiration.

    :param expiration: if a positive number, set an expiration on the cache, defaults to 0
    :type expiration: number, optional
    :param typed: If typed is set to true, function arguments of different types will be cached separately, defaults to False
    :type typed: bool, optional
    :param get_ts_func: The function to be used in order to get the current time, defaults to time.time
    :type get_ts_func: callable, optional
    :param log_recalculation: Whether or not to log cache misses, defaults to False
    :type log_recalculation: bool, optional
    :param ignored_keywords: Arguments to ignore when caculating item key, defaults to None
    :type ignored_keywords: iterable, optional
    :param key_func: The function to use in order to create the item key, defaults to functools._make_key
    :type key_func: callable, optional
    """

    def deco(func):
        return _TimeCache(
            func=func,
            expiration=expiration,
            typed=typed,
            get_ts_func=get_ts_func,
            log_recalculation=log_recalculation,
            ignored_keywords=ignored_keywords,
            key_func=key_func)

    return deco


timecache.__doc__ = _TimeCache.__doc__


@parametrizeable_decorator
def locking_cache(func=None, typed=False, log_recalculation=False, ignored_keywords=False):
    """
    A syntactic sugar for a locking cache without time expiration.

    :param typed: If typed is set to true, function arguments of different types will be cached separately, defaults to False
    :type typed: bool, optional
    :param log_recalculation: Whether or not to log cache misses, defaults to False
    :type log_recalculation: bool, optional
    :param ignored_keywords: Arguments to ignore when caculating item key, defaults to None
    :type ignored_keywords: iterable, optional
    """

    return timecache(typed=typed, log_recalculation=log_recalculation, ignored_keywords=ignored_keywords)(func)


class cached_property(object):
    """
    A property whose value is computed only once.

    :param function: Function to decorate, defaults to None
    :type function: function, optional
    :param locking: Lock cache access to make thread-safe, defaults to True
    :type locking: bool, optional
    :param safe: See `easypy.properties.safe_property`, defaults to True
    :type safe: bool, optional
    """

    locks_lock = RLock()
    LOCKS_KEY = '__cached_properties_locks'

    def __init__(self, function=None, locking=True, safe=True):
        self._function = function
        self._locking = locking
        self._safe = safe

    def __get__(self, obj, _=None):
        if obj is None:
            return self
        func_name = self._function.__name__
        with ExitStack() as stack:
            if self._locking:
                if not hasattr(obj, self.LOCKS_KEY):  # double synchronisation strategy to prevent more expensive lock entrance
                    with self.locks_lock:
                        if not hasattr(obj, self.LOCKS_KEY):
                            setattr(obj, self.LOCKS_KEY, defaultdict(RLock))
                stack.enter_context(getattr(obj, self.LOCKS_KEY)[func_name])

                try:
                    # another thread has cached the value by the time we acquired the lock
                    return getattr(obj, "_cached_%s" % func_name)
                except AttributeError:
                    pass

            try:
                value = self._function(obj)
            except AttributeError:
                if self._safe:
                    _, exc, tb = sys.exc_info()
                    raise RuntimeError("Attribute error within a property (%s)" % exc).with_traceback(tb)
                else:
                    raise

            setattr(obj, "_cached_%s" % func_name, value)
            setattr(obj, func_name, value)
            return value

    def __call__(self, func):
        self._function = func
        return self



adjectives = [
    "able", "abnormal", "admiring", "adoring", "adventurous",
    "affectionate", "agile", "agitated", "agreeable", "alert",
    "amazing", "ambitious", "amiable", "amusing", "analytical",
    "angelic", "angry", "apathetic", "apprehensive", "ardent",
    "artificial", "artistic", "assertive", "attentive", "average",
    "awesome", "awful", "balanced", "beautiful", "beneficent",
    "blissful", "blue", "blunt", "boisterous", "boring", "brave",
    "bright", "brilliant", "buff", "callous", "candid",
    "cantankerous", "capable", "careful", "careless", "caustic",
    "cautious", "charming", "cheerful", "chic", "childish",
    "childlike", "churlish", "circumspect", "civil", "clean",
    "clever", "clumsy", "cocky", "coherent", "cold", "compassionate",
    "competent", "composed", "conceited", "condescending",
    "confident", "confused", "conscientious", "considerate",
    "content", "cool", "cooperative", "cordial", "courageous",
    "cowardly", "crabby", "crafty", "cranky", "crass", "critical",
    "cruel", "curious", "cynical", "dainty", "dazzling", "decisive",
    "deep", "deferential", "deft", "delicate", "delightful",
    "demonic", "demure", "dependent", "depressed", "determined",
    "devoted", "dextrous", "diligent", "direct", "dirty",
    "disagreeable", "discerning", "discreet", "disruptive", "distant",
    "distracted", "distraught", "distrustful", "dowdy", "dramatic",
    "dreamy", "dreary", "drowsy", "drugged", "drunk", "dull",
    "dutiful", "eager", "earnest", "ecstatic", "efficient",
    "egotistical", "elastic", "elated", "elegant", "elfin",
    "eloquent", "emotional", "energetic", "enterprising",
    "enthusiastic", "epic", "evasive", "exacting", "excellent",
    "excitable", "experienced", "fabulous", "fastidious", "ferocious",
    "fervent", "festive", "fiery", "flabby", "flaky", "flamboyant",
    "flashy", "focused", "frank", "friendly", "frosty", "funny",
    "fussy", "gallant", "generous", "gentle", "gifted", "gloomy",
    "glutinous", "good", "goofy", "gracious", "grave", "great",
    "groggy", "grouchy", "guarded", "happy", "hardcore", "hateful",
    "hearty", "helpful", "hesitant", "heuristic", "hopeful", "hungry",
    "hypercritical", "hysterical", "idiotic", "idle", "illogical",
    "imaginative", "immature", "immodest", "impatient",
    "imperturbable", "impetuous", "impractical", "impressionable",
    "impressive", "impulsive", "inactive", "incisive", "incompetent",
    "inconsiderate", "inconsistent", "indefatigable", "independent",
    "indiscreet", "indolent", "industrious", "inexperienced",
    "infallible", "insensitive", "inspiring", "intelligent",
    "interesting", "intolerant", "inventive", "irascible",
    "irritable", "irritating", "jocular", "jolly", "jovial", "joyous",
    "judgmental", "keen", "kind", "lame", "laughing", "lazy", "lean",
    "leery", "lethargic", "listless", "lithe", "lively", "local",
    "logical", "lovable", "lovely", "loving", "lucid", "maternal",
    "mature", "mean", "meddlesome", "mercurial", "methodical",
    "meticulous", "mild", "miserable", "modest", "moronic", "morose",
    "motivated", "musical", "musing", "mystifying", "naive", "nasty",
    "natural", "naughty", "negative", "nervous", "nifty", "noisy",
    "normal", "nostalgic", "nosy", "numb", "objective", "obliging",
    "obnoxious", "optimistic", "orderly", "ostentatious", "outgoing",
    "outspoken", "passionate", "passive", "paternal", "paternalistic",
    "patient", "peaceful", "pedantic", "peevish", "pensive",
    "persevering", "persnickety", "petulant", "picky", "plain",
    "playful", "pleasant", "plucky", "polite", "popular", "positive",
    "powerful", "practical", "prejudiced", "pretty", "priceless",
    "proficient", "proud", "provocative", "prudent", "punctual",
    "quarrelsome", "querulous", "quick", "quiet", "quirky",
    "quizzical", "realistic", "reassuring", "reclusive", "relaxed",
    "reliable", "reluctant", "resentful", "reserved", "resigned",
    "resourceful", "respected", "respectful", "responsible",
    "restless", "revered", "reverent", "ridiculous", "romantic",
    "sad", "sassy", "saucy", "sedate", "selfish", "sensible",
    "sensitive", "sentimental", "serene", "serious", "sharp",
    "shrewd", "shy", "silly", "sincere", "sleepy", "slight", "sloppy",
    "slothful", "slovenly", "slow", "smart", "snazzy", "sneering",
    "snobby", "sober", "somber", "sophisticated", "soulful",
    "soulless", "sour", "spirited", "spiteful", "stable", "staid",
    "steady", "stern", "stoic", "striking", "strong", "stupefied",
    "stupid", "sturdy", "subtle", "sulky", "sullen", "supercilious",
    "superficial", "surly", "suspicious", "sweet", "tactful",
    "tactless", "talented", "tender", "testy", "thinking", "thirsty",
    "thoughtful", "thoughtless", "timid", "tired", "tolerant",
    "touchy", "tranquil", "trusting", "ugly", "unaffected",
    "unbalanced", "uncertain", "uncooperative", "undependable",
    "unemotional", "unfriendly", "unguarded", "unhelpful",
    "unimaginative", "unmotivated", "unpleasant", "unpopular",
    "unreliable", "unruffled", "unsophisticated", "unsure",
    "unthinking", "unwilling", "upbeat", "venal", "versatile",
    "vibrant", "vigilant", "vigorous", "volcanic", "vulnerable",
    "warm", "warmhearted", "wary", "watchful", "weak", "willing",
    "wizardly", "wonderful", "youthful", "zealous"
]

creatures = [
    "accentor", "acouchi", "addax", "agama", "agouti", "albatross",
    "alligator", "alpaca", "anaconda", "angelfish", "anhinga", "anoa",
    "anole", "ant", "anteater", "antelope", "archerfish", "armadillo",
    "arowana", "auk", "avocet", "axolotl", "babirusa", "baboon",
    "badger", "bandicoot", "banshee", "banteng", "barbet", "basilisk",
    "bat", "batfish", "beaver", "bee", "beetle", "betta", "bettong",
    "bigeyes", "bilby", "binturong", "bird", "bison", "bitterling",
    "bittern", "blackdevil", "bluebird", "boa", "bongo", "booby",
    "bowerbird", "broadbill", "bulbul", "bunting", "bushbaby",
    "bushbaby", "bushshrike", "bustard", "butterfly", "buzzard",
    "caecilian", "camel", "capybara", "caracal", "caracara",
    "cardinal", "cardinalfish", "caribou", "caribou", "carp",
    "cassowary", "cat", "caterpillar", "catfish", "cattle", "cattle",
    "cattle", "cavy", "centaur", "centipede", "cerberus", "chameleon",
    "chamois", "charybdis", "chat", "cheetah", "chicken", "chimaera",
    "chimera", "chimpanzee", "chinchilla", "chipmunk", "chough",
    "chuckwalla", "cicada", "cichlid", "civet", "clam", "clownfish",
    "coati", "cobra", "cockatiel", "cockatoo", "cockroach", "colugo",
    "conch", "coot", "copperhead", "cormorant", "cotinga", "cougar",
    "cougar", "courser", "coyote", "coypu", "coypu", "crab", "crane",
    "crayfish", "cricket", "crocodile", "crow", "cuckoo", "culpeo",
    "curlew", "cuscus", "cuttlefish", "cynocephalus", "dartfish",
    "deer", "degu", "demon", "dhole", "dibbler", "dikkop", "dikkop",
    "dikkop", "dimetrodon", "dingo", "dinosaur", "dipper", "discus",
    "dodo", "dog", "dolphin", "donkey", "doppelganger", "dormouse",
    "dove", "dragonfly", "drongo", "duck", "dugong", "duiker",
    "dunnart", "eagle", "echidna", "echidna", "eel", "egret", "eider",
    "elephant", "elk", "elk", "emu", "fairy", "falcon", "ferret",
    "finch", "fish", "fisher", "flamingo", "flea", "flowerpecker",
    "fly", "fossa", "fox", "frigatebird", "frog", "frogmouth",
    "fulmar", "gannet", "gar", "gaur", "gazelle", "gecko", "gerbil",
    "gibbon", "giraffe", "goat", "goatfish", "goblin", "goldfish",
    "golem", "goose", "gopher", "gopher", "goral", "gorilla",
    "gourami", "grackle", "grasshopper", "grebe", "griffin", "grison",
    "groundhog", "groundhog", "grouse", "guanaco", "gull", "gull",
    "gundi", "hamster", "harrier", "hartebeest", "hawk", "hawkodile",
    "hawkodile", "hedgehog", "helmetshrike", "heron", "hippopotamus",
    "honeyeater", "hornbill", "hornet", "horse", "hoverfly",
    "hummingbird", "hutia", "hyena", "hyrax", "ibex", "ibis",
    "icterid", "iguana", "impala", "insect", "jacana", "jack",
    "jackal", "jaguar", "jaguarundi", "jay", "jellyfish", "jerboa",
    "kangaroo", "kerodon", "kestrel", "kingbird", "kingfisher",
    "kinkajou", "kite", "kitten", "kiwi", "klipspringer", "knifefish",
    "koala", "kodkod", "kookaburra", "kowari", "kudu", "kultarr",
    "ladybug", "lamprey", "lapwing", "leech", "lemming", "lemur",
    "leopard", "leprechaun", "liger", "lion", "lionfish", "lizard",
    "llama", "loach", "lobster", "longspur", "loon", "loris", "lory",
    "lovebird", "lynx", "lyrebird", "macaw", "mamba", "mammoth",
    "manakin", "manatee", "mandrill", "manticore", "mara", "margay",
    "markhor", "marlin", "marmot", "marten", "mastodon", "meadowlark",
    "meerkat", "megaloceros", "megapode", "mermaids", "millipede",
    "mink", "minnow", "minotaur", "mockingbird", "mole", "mongoose",
    "monkey", "moorhen", "moorhen", "moose", "mosasaur", "mosquito",
    "moth", "mothman", "motmot", "mouse", "mousebird", "mudpuppy",
    "mudskipper", "mullet", "muntjac", "muskox", "muskrat", "mutants",
    "narwhal", "nautilus", "needlefish", "newt", "nighthawk",
    "nightjar", "numbat", "nuthatch", "ocelot", "octopus", "okapi",
    "olingo", "onager", "opossum", "orangutan", "orca", "oriole",
    "orthros", "oryx", "osprey", "ostrich", "otter", "ovenbird",
    "owl", "oyster", "paca", "paddlefish", "pademelon", "pangolin",
    "panther", "parakeet", "parakeet", "parrot", "parrotfish",
    "peafowl", "peafowl", "peccary", "pelican", "penguin", "pheasant",
    "pig", "pigeon", "pika", "pike", "piranha", "pitohui", "platypus",
    "plover", "pony", "porcupine", "porpoise", "possum", "potoo",
    "potoroo", "potto", "pronghorn", "ptarmigan", "pterosaur", "pudu",
    "puffin", "puma", "puppy", "python", "quagga", "quail", "quetzal",
    "quokka", "quoll", "rabbit", "rabbit", "raccoon", "rail",
    "rainbowfish", "rat", "rattlesnake", "raven", "rhea",
    "rhinoceros", "ringtail", "roadrunner", "robin", "rockfish",
    "roller", "rook", "sailfish", "salamander", "salmon",
    "sandgrouse", "sawfish", "scorpion", "scylla", "seadragon",
    "seahorse", "seal", "serval", "shapeshifter", "shark",
    "shearwater", "sheep", "sheep", "shelduck", "shoebill", "shrew",
    "shrimp", "shrimp", "siren", "skate", "skimmer", "skink", "skua",
    "skunk", "sloth", "slug", "snail", "snake", "snipe", "solenodon",
    "songbird", "sparrow", "spider", "sponge", "spoonbill",
    "springhare", "squid", "squirrel", "starfish", "starling",
    "stingray", "stonefish", "stork", "sturgeon", "sunbird",
    "surgeonfish", "swallow", "swamphen", "swan", "swift",
    "swordfish", "tadpole", "takin", "tamandua", "tamarin", "tanager",
    "tapaculo", "tapir", "tarantula", "tarpon", "tarsier", "tayra",
    "tenrec", "termite", "tern", "tetra", "thornbill", "thrasher",
    "thrush", "thunderbird", "tiger", "toad", "toadfish", "tortoise",
    "toucan", "treeshrew", "trogon", "trout", "trumpeter", "tuatara",
    "tuna", "turaco", "turkey", "turtle", "typhon", "unicorn",
    "urchin", "vampire", "vanga", "vaquita", "vicuna", "viper",
    "viscacha", "vole", "vulture", "wader", "wagtail", "wallaby",
    "walleye", "walrus", "warbler", "warthog", "wasp", "waxwing",
    "weasel", "weaver", "wendigo", "werewolf", "whale", "whistler",
    "whydah", "wigeon", "wildcat", "wildebeest", "wolf",
    "wolverine", "wombat", "woodcock", "woodpecker", "woodswallow",
    "worm", "wren", "xenopoecilus", "xenops", "xerus", "yak",
    "yellowjacket", "zebra", "zebu", "zombie"
]



"""
This module helps you generate deprecation warnings
"""

from functools import wraps, partial
import warnings


def deprecated(func=None, message=None):
    if not callable(func):
        return partial(deprecated, message=func)
    message = (" " + message) if message else ""
    message = "Hey! '%s' is deprecated!%s" % (func.__name__, message)

    @wraps(func)
    def inner(*args, **kwargs):
        warnings.warn(message, DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)
    return inner


def deprecated_arguments(**argmap):
    """
    Renames arguments while emitting deprecation warning::

        @deprecated_arguments(old_name='new_name')
        def func(new_name):
            # ...

        func(old_name='value meant for new name')
    """

    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            deprecation_warnings = []
            for name, map_to in argmap.items():
                try:
                    value = kwargs.pop(name)
                except KeyError:
                    pass  # deprecated argument was not used
                else:
                    if map_to in kwargs:
                        raise TypeError("%s is deprecated for %s - can't use both in %s()" % (
                                        name, map_to, func.__name__))
                    deprecation_warnings.append('%s is deprecated - use %s instead' % (name, map_to))
                    kwargs[map_to] = value

            if deprecation_warnings:
                message = 'Hey! In %s, %s' % (func.__name__, ', '.join(deprecation_warnings))
                warnings.warn(message, DeprecationWarning, stacklevel=2)

            return func(*args, **kwargs)
        return inner
    return wrapper



import re
import time
from datetime import datetime
from queue import PriorityQueue
from .colors import uncolored


TIMESTAMP_PATTERN = "%Y-%m-%d %H:%M:%S"


def to_timestamp(t):
    return "-".center(19) if t is None else time.strftime(TIMESTAMP_PATTERN, time.localtime(t))


YEAR = time.strftime("%Y")
MONTH = time.strftime("%m")
DAY = time.strftime("%d")

TIMESTAMP_GETTERS = [

    # 01:21:27
    (re.compile(r"^(\d+:\d+:\d+)"),
     lambda ts: time.mktime(time.strptime("%s-%s-%s %s" % (YEAR, MONTH, DAY, ts), "%Y-%m-%d %H:%M:%S"))),

    # Apr 6 17:13:40
    (re.compile(r"^(\w{3} +\d+ +\d+:\d+:\d+)"),
     lambda ts: time.mktime(time.strptime("%s %s" % (YEAR, ts), "%Y %b %d %H:%M:%S"))),

    # 2018-12-15T02:11:06+0200
    (re.compile(r"^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{4})"),
     lambda ts: datetime.strptime(ts, '%Y-%m-%dT%H:%M:%S%z').timestamp()),

    # 2018-12-15T02:11:06.123456+02:00
    # 2019-10-09T10:58:45,929228489+03:00
    (re.compile(r"^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})[.,](\d{6})\d*(\+\d{2}):(\d{2})"),
     lambda *args: datetime.strptime("{}.{}{}{}".format(*args), '%Y-%m-%dT%H:%M:%S.%f%z').timestamp()),

    # 2018-04-06 17:13:40,955
    # 2018-04-23 04:48:11,811|
    (re.compile(r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}),(\d{3})[| ]"),
     lambda ts, ms: time.mktime(time.strptime(ts, "%Y-%m-%d %H:%M:%S")) + float(ms) / 1000),

    # 2018-04-06 17:13:40
    # 2018-04-06 17:13:40.955356
    # [2018/04/06 17:13:40
    # [2018/04/06 17:13:40.955356
    (re.compile(r"^\[?(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})(?:\.(\d{6}))?"),
     lambda ts, ms: time.mktime(time.strptime(ts.replace("/", "-"), "%Y-%m-%d %H:%M:%S")) + float(ms or 0) / 1000000),

    # for strace logs
    # 16255 15:08:52.554223
    (re.compile(r"\d+ (\d{2}:\d{2}:\d{2}).(\d{6})"),
     lambda ts, ms: time.mktime(time.strptime("%s-%s-%s %s" % (YEAR, MONTH, DAY, ts), "%Y-%m-%d %H:%M:%S")) + float(ms) / 1000000),
]


class TimestampedStream(object):

    def __init__(self, stream, prefix="> "):
        self.name = stream.name if hasattr(stream, "name") else repr(stream)
        self.stream = iter(stream)
        self.prefix = prefix
        self.filler = " "*len(self.prefix)
        self._untimestamp = None

    def __gt__(self, other):
        return id(self) > id(other)

    def get_next(self):
        "Get next line in the stream and the stream itself, or None if stream ended"
        try:
            line = next(self.stream)
        except StopIteration:
            return
        else:
            ts = self.get_timestamp(uncolored(line))
            return ts, (self.prefix if ts else self.filler) + line, self

    def get_timestamp(self, line):
        """
        Find the timestamp if exists, and return as a float
        """
        if not line.startswith(" "):
            if not self._untimestamp:
                for regex, converter in TIMESTAMP_GETTERS:
                    match = regex.search(line)
                    if match:
                        # cache the regex and conversion funcs for later
                        self._untimestamp = converter, regex
                        break
            else:
                converter, regex = self._untimestamp
                match = regex.search(line)
            if match:
                return converter(*match.groups())
        return 0


def iter_zipped_logs(*log_streams, prefix="> ", show_intervals=None):
    """
    Line iterator that merges lines from different log streams based on their timestamp.
    Timestamp patterns are found in the TIMESTAMP_GETTERS list in this module.

    :param prefix: Prepend this prefix to each line where a timestamp was identified
    :param show_intervals: `s` or `ms` - Prepend the duration since the previous log line
    """

    # A sorted queue of (timestamp, stream) tuples (lowest-timestamp first)
    streams = PriorityQueue()
    stream_names = []
    for i, stream in enumerate(log_streams):
        if not isinstance(stream, tuple):
            tstream = TimestampedStream(stream, prefix)
        else:
            tstream = TimestampedStream(*stream)

        n = tstream.get_next()
        if n:
            stream_names.append(tstream.name)
            streams.put(n)

    last_ts = None
    if show_intervals:
        from easypy.units import Duration

        def formatted(line, current_ts, last_ts):
            fmt = "{:>7}{}"
            if (current_ts and last_ts):
                return fmt.format(Duration(current_ts - last_ts).render(show_intervals), line)
            else:
                return fmt.format("", line)
    else:
        def formatted(line, current_ts, last_ts):
            return line

    while not streams.empty():
        current_ts, line, stream = streams.get()
        yield formatted(line, current_ts, last_ts)
        last_ts = current_ts
        while True:
            n = stream.get_next()
            if not n:
                break   # stream ended
            ts, line, stream = n
            if ts and ts > current_ts:
                streams.put((ts, line, stream))
                break   # timestamp advanced
            yield formatted(line, ts, last_ts)
            if ts:
                last_ts = ts


def main():
    import sys
    import argparse
    parser = argparse.ArgumentParser(description='ZipLog - merge logs by timestamps')
    parser.add_argument(
        'logs', metavar='N', type=str, nargs='+',
        help='Log files; Use "-" for STDIN')
    parser.add_argument(
        '-i', '--interval', dest='interval', default=None,
        help="Show interval by seconds (s), or milliseconds (ms)")
    parser.add_argument(
        '-p', '--prefix', dest='prefix', default="> ",
        help="A prefix to prepend to timestamped lines")
    ns = parser.parse_args(sys.argv[1:])

    files = [sys.stdin if f == "-" else open(f) for f in ns.logs]
    try:
        for line in iter_zipped_logs(*files, show_intervals=ns.interval, prefix=ns.prefix):
            print(line, end="")
    except BrokenPipeError:
        pass


if __name__ == "__main__":
    main()



from easypy.bunch import Bunch, bunchify


def test_bunch_recursion():
    x = Bunch(a='a', b="b", d=Bunch(x="axe", y="why"))
    x.d.x = x
    x.d.y = x.b
    print(x)


def test_bunchify():
    x = bunchify(dict(a=[dict(b=5), 9, (1, 2)], c=8))
    assert x.a[0].b == 5
    assert x.a[1] == 9
    assert isinstance(x.a[2], tuple)
    assert x.c == 8
    assert x.pop("c") == 8



from easypy.exceptions import TException

# TODO: some methods like fromkeys/update can insert keys which are not strings - and that causes problems.
#       need to either convert keys to strings (like in _convert), or raise ValueError.
class Bunch(dict):

    __slots__ = ("__stop_recursing__",)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            if name[0] == "_" and name[1:].isdigit():
                return self[name[1:]]
            raise AttributeError("%s has no attribute %r" % (self.__class__, name))

    def __getitem__(self, key):
        try:
            return super(Bunch, self).__getitem__(key)
        except KeyError:
            from numbers import Integral
            if isinstance(key, Integral):
                return self[str(key)]
            raise

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        try:
            del self[name]
        except KeyError:
            raise AttributeError("%s has no attribute %r" % (self.__class__, name))

    def __getstate__(self):
        return self

    def __setstate__(self, dict):
        self.update(dict)

    def __repr__(self):
        if getattr(self, "__stop_recursing__", False):
            items = sorted("%s" % k for k in self if isinstance(k, str) and not k.startswith("__"))
            attrs = ", ".join(items)
        else:
            dict.__setattr__(self, "__stop_recursing__", True)
            try:
                attrs = self.render()
            finally:
                dict.__delattr__(self, "__stop_recursing__")
        return "%s(%s)" % (self.__class__.__name__, attrs)

    def render(self):
        items = sorted("%s=%r" % (k, v) for k, v in self.items() if isinstance(k, str) and not k.startswith("__"))
        return ", ".join(items)

    def _repr_pretty_(self, *args, **kwargs):
        from easypy.humanize import ipython_mapping_repr
        return ipython_mapping_repr(self, *args, **kwargs)

    def to_dict(self):
        return unbunchify(self)

    def to_json(self):
        import json
        return json.dumps(self.to_dict())

    def to_yaml(self):
        import yaml
        return yaml.dump(self.to_dict())

    def copy(self, deep=False):
        if deep:
            return _convert(self, self.__class__)
        else:
            return self.__class__(self)

    @classmethod
    def from_dict(cls, d):
        return _convert(d, cls)

    @classmethod
    def from_json(cls, d):
        import json
        return cls.from_dict(json.loads(d))

    @classmethod
    def from_yaml(cls, d):
        import yaml
        return cls.from_dict(yaml.load(d))

    @classmethod
    def from_xml(cls, d):
        import xmltodict
        return cls.from_dict(xmltodict.parse(d))

    def __dir__(self):
        members = set(k for k in self if isinstance(k, str) and (k[0] == "_" or k.replace("_", "").isalnum()))
        members.update(dict.__dir__(self))
        return sorted(members)

    def without(self, *keys):
        "Return a shallow copy of the bunch without the specified keys"
        return Bunch((k, v) for k, v in self.items() if k not in keys)

    def but_with(self, **kw):
        "Return a shallow copy of the bunch with the specified keys"
        return Bunch(self, **kw)


def _convert(d, typ):
    if isinstance(d, dict):
        return typ(dict((str(k), _convert(v, typ)) for k, v in d.items()))
    elif isinstance(d, (tuple, list, set)):
        return type(d)(_convert(e, typ) for e in  d)
    else:
        return d


def unbunchify(d):
    """Recursively convert Bunches in `d` to a regular dicts."""
    return _convert(d, dict)


def bunchify(d=None, **kw):
    """Recursively convert dicts in `d` to Bunches.
    If `kw` given, recursively convert dicts in it to Bunches and update `d` with it.
    If `d` is None, an empty Bunch is made."""
    d = _convert(d, Bunch) if d is not None else Bunch()
    if kw:
        d.update(bunchify(kw))
    return d



import pytest

from functools import wraps

from easypy.decorations import lazy_decorator
from easypy.decorations import ensure_same_defaults, DefaultsMismatch
from easypy.misc import kwargs_resilient


def test_kwargs_resilient():
    @kwargs_resilient
    def foo(a, b):
        return [a, b]

    assert foo(1, b=2, c=3, d=4) == [1, 2]

    @kwargs_resilient
    def bar(a, b, **kwargs):
        return [a, b, kwargs]

    assert bar(1, b=2, c=3, d=4) == [1, 2, {'c': 3, 'd': 4}]

    @kwargs_resilient(negligible='d')
    def baz(a, b):
        return [a, b]

    # Should only be neglect `d` - not to `c`
    with pytest.raises(TypeError):
        baz(1, b=2, c=3, d=4)
    assert baz(1, b=2, d=4) == [1, 2]

    @kwargs_resilient(negligible=['b', 'd'])
    def qux(a, b, **kwargs):
        return [a, b, kwargs]

    # Should be passing b because it's in the function signature
    # Should be passing c because it's not in `negligible`
    # Should not be passing d because it's in `negligible` and not in the function signature
    assert qux(1, b=2, c=3, d=4) == [1, 2, {'c': 3}]


def test_lazy_decorator_lambda():
    def add_to_result(num):
        def inner(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs) + num

            wrapper.__name__ = '%s + %s' % (func.__name__, num)

            return wrapper
        return inner

    class Foo:
        def __init__(self, num):
            self.num = num

        @lazy_decorator(lambda self: add_to_result(num=self.num))
        def foo(self):
            """foo doc"""
            return 1

    foo = Foo(10)
    assert foo.foo() == 11

    assert Foo.foo.__name__ == 'foo'
    assert foo.foo.__name__ == 'foo + 10'

    assert Foo.foo.__doc__ == foo.foo.__doc__ == 'foo doc'


def test_lazy_decorator_attribute():
    class Foo:
        def add_to_result(self, func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs) + self.num

            wrapper.__name__ = '%s + %s' % (func.__name__, self.num)

            return wrapper

        @lazy_decorator('add_to_result')
        def foo(self):
            """foo doc"""
            return 1

    foo = Foo()

    with pytest.raises(AttributeError):
        # We did not set foo.num yet, so the decorator will fail trying to set the name
        foo.foo

    foo.num = 10
    assert foo.foo() == 11
    assert foo.foo.__name__ == 'foo + 10'
    assert Foo.foo.__doc__ == foo.foo.__doc__ == 'foo doc'

    foo.num = 20
    assert foo.foo() == 21
    assert foo.foo.__name__ == 'foo + 20'


def test_lazy_decorator_with_timecache():
    from easypy.caching import timecache

    class Foo:
        def __init__(self):
            self.ts = 0
            self._counter = 0

        @property
        def timecache(self):
            return timecache(expiration=1, get_ts_func=lambda: self.ts)

        @lazy_decorator('timecache', cached=True)
        def inc(self):
            self._counter += 1
            return self._counter

        @lazy_decorator(lambda self: lambda method: method())
        @lazy_decorator('timecache', cached=True)
        def counter(self):
            return self._counter

    foo1 = Foo()
    foo2 = Foo()

    assert [foo1.inc(), foo2.inc()] == [1, 1]
    assert [foo1.inc(), foo2.inc()] == [1, 1]
    assert [foo1.counter, foo2.counter] == [1, 1]

    foo1.ts += 1
    assert [foo1.counter, foo2.counter] == [1, 1]
    assert [foo1.inc(), foo2.inc()] == [2, 1]
    assert [foo1.counter, foo2.counter] == [1, 1]
    foo2.ts += 1
    assert [foo1.inc(), foo2.inc()] == [2, 2]
    assert [foo1.counter, foo2.counter] == [1, 2]  # foo1 was not updated since last sync - only foo2


def test_ensure_same_defaults():
    def foo(a=1, b=2, c=3):
        return a, b, c


    @ensure_same_defaults(foo)
    def bar(a=1, b=2, c=3):
        return a, b, c

    # Test we did not change the actual function
    assert foo() == bar()
    assert foo(4, 5, 6) == bar(4, 5, 6)


    with pytest.raises(DefaultsMismatch) as exc:
        @ensure_same_defaults(foo)
        def baz(a=1, b=3, c=2):
            pass
    assert exc.value.param_names == ['b', 'c']


def test_ensure_same_defaults_skipping_params_with_no_default():
    @ensure_same_defaults(lambda a=1, b=2: ...)
    def foo(a, b=2):
        pass

    @ensure_same_defaults(lambda a, b=2: ...)
    def foo(a=1, b=2):
        pass

    @ensure_same_defaults(lambda a, b=2: ...)
    def foo(a=1, c=3):
        pass

    with pytest.raises(DefaultsMismatch) as exc:
        @ensure_same_defaults(lambda a, b=2: ...)
        def foo(a, b=4):
            pass
    assert exc.value.param_names == ['b']


def test_ensure_same_defaults_ignore():
    @ensure_same_defaults(lambda a=1, b=2: ..., ignore=('b',))
    def foo(a=1, b=3):
        pass

    with pytest.raises(DefaultsMismatch) as exc:
        @ensure_same_defaults(lambda a=1, b=2: ..., ignore=('b',))
        def foo(a=2, b=3):
            pass
    assert exc.value.param_names == ['a']



from unittest.mock import patch
import pytest
from time import sleep

from easypy.threadtree import get_thread_stacks, ThreadContexts
from easypy.concurrency import concurrent, MultiObject, MultiException


@pytest.yield_fixture(params=[True, False], ids=['concurrent', 'nonconcurrent'])
def concurrency_enabled_and_disabled(request):
    if request.param:  # concurrency enabled
        yield
    else:  # concurrency disabled
        from easypy.concurrency import disable, enable
        try:
            disable()
            yield
        finally:
            enable()


def test_thread_stacks():
    with concurrent(sleep, .1, threadname='sleep'):
        print(get_thread_stacks().render())


def test_thread_contexts_counters():
    TC = ThreadContexts(counters=('i', 'j'))
    assert TC.i == TC.j == 0

    with TC(i=1):
        def check1():
            assert TC.i == 1
            assert TC.j == 0

            with TC(i=1, j=1):
                def check2():
                    assert TC.i == 2
                    assert TC.j == 1
                with concurrent(check2):
                    pass

        with concurrent(check1):
            pass


def test_thread_contexts_counters_multiobject():
    TC = ThreadContexts(counters=('i',))
    assert TC.i == 0

    print("---")

    @TC(i=True)
    def test(n):
        print(n, TC._context_data)
        sleep(.1)
        return TC.i

    test(0)
    ret = MultiObject(range(10)).call(test)
    assert set(ret) == {1}


def test_thread_context_stacks():
    TC = ThreadContexts(stacks=('i', 'j'))
    assert TC.i == TC.j == []

    with TC(i='a'):
        def check1():
            assert TC.i == ['a']
            assert TC.j == []

            with TC(i='i', j='j'):
                def check2():
                    assert TC.i == ['a', 'i']
                    assert TC.j == ['j']
                with concurrent(check2):
                    pass

        with concurrent(check1):
            pass


def test_multiobject_0():
    x = MultiObject([]).foo()
    assert len(x) == 0
    assert x.__class__.CONCESTOR is object


def test_multiobject_1():
    m = MultiObject(range(10))

    def mul(a, b, *c):
        return a * b + sum(c)

    assert sum(m.call(mul, 2)) == 90
    assert sum(m.call(mul, b=10)) == 450
    assert sum(m.call(mul, 1, 1, 1)) == 65

    assert m.filter(None).T == (1, 2, 3, 4, 5, 6, 7, 8, 9)
    assert sum(m.denominator) == 10
    with pytest.raises(MultiException) as info:
        m.call(lambda i: 1 / (i % 2))

    assert info.value.count == 5
    assert info.value.common_type == ZeroDivisionError
    assert not info.value.complete


def test_multiobject_exceptions():

    assert MultiException[ValueError] is MultiException[ValueError]
    assert issubclass(MultiException[UnicodeDecodeError], MultiException[UnicodeError])
    assert issubclass(MultiException[UnicodeDecodeError], MultiException[ValueError])

    with pytest.raises(AssertionError):
        MultiException[0]

    with pytest.raises(MultiException):
        MultiObject(range(5)).call(lambda n: 1 / n)

    with pytest.raises(MultiException[Exception]):
        MultiObject(range(5)).call(lambda n: 1 / n)

    with pytest.raises(MultiException[ZeroDivisionError]):
        MultiObject(range(5)).call(lambda n: 1 / n)

    try:
        MultiObject(range(5)).call(lambda n: 1 / n)
    except MultiException[ValueError] as exc:
        assert False
    except MultiException[ZeroDivisionError] as exc:
        assert len(exc.actual) == 1
        assert isinstance(exc.one, ZeroDivisionError)
    else:
        assert False

    with pytest.raises(MultiException[ArithmeticError]):
        try:
            MultiObject(range(5)).call(lambda n: 1 / n)
        except ZeroDivisionError:
            assert False  # shouldn't be here
        except MultiException[ValueError]:
            assert False  # shouldn't be here


class ExceptionForPicklingTest(ArithmeticError):
    pass


def test_multiexception_pickling():
    import pickle
    import multiprocessing

    def throw(n):
        if not n:
            raise ExceptionForPicklingTest(n)

    def fail_and_dump(queue):
        try:
            MultiObject(range(5)).call(throw)
        except MultiException[ArithmeticError] as exc:
            p = pickle.dumps(exc)
            queue.put_nowait(p)

    queue = multiprocessing.Queue(1)
    process = multiprocessing.Process(target=fail_and_dump, args=(queue,))
    process.start()
    process.join()
    p = queue.get_nowait()

    exc = pickle.loads(p)
    assert isinstance(exc, MultiException[ExceptionForPicklingTest])
    assert exc.common_type is ExceptionForPicklingTest
    assert exc.exceptions[0].args == (0,)
    assert exc.exceptions[1:] == [None] * 4


def test_multiobject_concurrent_find_found():
    m = MultiObject(range(10))
    from time import sleep
    ret = m.concurrent_find(lambda n: sleep(n / 10) or n)  # n==0 is not nonzero, so it's not eligible
    assert ret == 1


def test_multiobject_concurrent_find_not_found():
    m = MultiObject(range(10))
    ret = m.concurrent_find(lambda n: n < 0)
    assert ret is False

    m = MultiObject([0] * 5)
    ret = m.concurrent_find(lambda n: n)
    assert ret == 0


def test_multiobject_concurrent_find_proper_shutdown():
    executed = []
    m = MultiObject(range(10), workers=1)
    ret = m.concurrent_find(lambda n: [print(n) or executed.append(n) or sleep(.01)])
    assert ret
    sleep(1)  # wait for potential stragglers
    assert max(executed) <= 2


def test_multiobject_zip_with():
    m = MultiObject(range(4))

    with pytest.raises(AssertionError):
        m.zip_with(range(3), range(5))  # too few objects

    m.zip_with(range(5), range(6))  # too many objects

    ret = m.zip_with(range(1, 5)).call(lambda a, b: a + b).T
    assert ret == (1, 3, 5, 7)


def test_multiobject_enumerate():
    m = MultiObject(range(5), log_ctx="abcd")

    def check(i, j):
        assert i == j + 1

    e = m.enumerate(1)
    assert e._log_ctx == tuple("abcd")
    e.call(check)


def test_multiobject_logging():
    m = MultiObject(range(4), log_ctx="abcd", initial_log_interval=0.1)

    def check(i):
        sleep(.2)

    # we'll mock the logger so we can ensure it logged
    with patch("easypy.concurrency._logger") as _logger:
        m.call(check)

    args_list = (c[0] for c in _logger.info.call_args_list)
    for args in args_list:
        assert "test_multiobject_logging.<locals>.check" == args[2]
        assert "easypy/tests/test_concurrency.py" in args[4]


def test_multiobject_types():
    assert isinstance(MultiObject(range(5)), MultiObject[int])
    assert not isinstance(MultiObject(range(5)), MultiObject[str])

    class A(): ...
    class B(A): ...

    assert issubclass(MultiObject[A], MultiObject)
    assert not issubclass(MultiObject[A], A)
    assert issubclass(MultiObject[B], MultiObject[A])
    assert not issubclass(MultiObject[A], MultiObject[B])

    assert isinstance(MultiObject([B()]), MultiObject[A])
    assert not isinstance(MultiObject([A()]), MultiObject[B])
    assert isinstance(MultiObject[A]([B()]), MultiObject[A])
    assert isinstance(MultiObject[A]([B()]), MultiObject[B])
    assert isinstance(MultiObject[int](range(5)), MultiObject[int])

    with pytest.raises(TypeError):
        assert MultiObject[str](range(5))

    assert isinstance(MultiObject[str]("123").call(int), MultiObject[int])


def test_multiobject_namedtuples():
    from collections import namedtuple

    class Something(namedtuple("Something", "a b")):
        pass

    def ensure_not_expanded(something):
        # This will probably fail before these asserts
        assert hasattr(something, 'a')
        assert hasattr(something, 'b')

    objects = [Something(1, 2), Something(2, 3), Something(3, 4)]
    MultiObject(objects).call(ensure_not_expanded)


@pytest.mark.usefixtures('concurrency_enabled_and_disabled')
def test_multiexception_api():
    with pytest.raises(MultiException) as exc:
        MultiObject([0, 5]).call(lambda i: 10 // i)

    failed, sucsessful = exc.value.futures

    assert failed.done()
    with pytest.raises(ZeroDivisionError):
        failed.result()
    assert isinstance(failed.exception(), ZeroDivisionError)

    assert sucsessful.done()
    assert sucsessful.result() == 2
    assert sucsessful.exception() is None


def test_multiexception_types():

    class OK(Exception):
        pass

    class BAD(object):
        pass

    class OKBAD(OK, BAD):
        pass

    with pytest.raises(AssertionError):
        MultiException[BAD]

    def raise_it(typ):
        raise typ()

    with pytest.raises(MultiException[OK]):
        MultiObject([OK]).call(raise_it)

    with pytest.raises(MultiException[OKBAD]):
        MultiObject([OKBAD]).call(raise_it)

    with pytest.raises(MultiException[OK]):
        MultiObject([OKBAD]).call(raise_it)


@pytest.mark.parametrize('throw', [False, True])
def test_concurrent_done_status(throw):
    from threading import Event

    continue_func = Event()

    def func():
        continue_func.wait()
        if throw:
            raise Exception()

    with concurrent(func, throw=False) as c:
        assert not c.done()
        continue_func.set()
        sleep(0.1)
        assert c.done()
    assert c.done()


def test_concurrent_real_thread():
    from easypy.concurrency import IS_GEVENT
    from gevent.monkey import get_original
    import logging

    sleep = get_original("time", "sleep")
    current_thread = get_original("threading", "get_ident")
    main_thread = current_thread()

    ran = 0

    def log_and_sleep():
        nonlocal ran
        logging.info("test")
        sleep(.1)
        ran += 1
        return current_thread()

    if IS_GEVENT:
        before = ran
        with concurrent(log_and_sleep, real_thread_no_greenlet=True) as c:
            pass
        result = c.result()
        assert ran == before + 1
        assert main_thread != result

        before = ran
        with concurrent(log_and_sleep) as c:
            pass
        result = c.result()
        assert ran == before + 1
        assert main_thread == result
    else:
        before = ran
        with concurrent(log_and_sleep, real_thread_no_greenlet=True) as c:
            pass
        result = c.result()
        assert ran == before + 1
        assert main_thread != result

        before = ran
        with concurrent(log_and_sleep) as c:
            pass
        result = c.result()
        assert ran == before + 1
        assert main_thread != result

    assert ran



import pytest
from contextlib import contextmanager
from io import StringIO
from logging import getLogger, Formatter, StreamHandler
from easypy.colors import uncolored
logger = getLogger(__name__)


@pytest.yield_fixture
def get_log():
    level = logger.root.level
    stream = StringIO()
    handler = StreamHandler(stream)
    handler.setFormatter(Formatter(fmt="%(message)s"))
    logger.root.addHandler(handler)
    logger.root.setLevel(0)

    def get():
        return uncolored(stream.getvalue())
    yield get

    logger.root.setLevel(level)
    logger.root.removeHandler(handler)


def test_indent_around_generator(get_log):

    @logger.indented("hey")
    def gen():
        logger.info("000")
        yield 1
        yield 2

    for i in gen():
        logger.info("%03d", i)
        break

    assert get_log() == "hey\n000\n001\nDONE in no-time (hey)\n"


def test_indent_around_function(get_log):

    @logger.indented("hey")
    def foo():
        logger.info("001")

    foo()

    assert get_log() == "hey\n001\nDONE in no-time (hey)\n"


def test_indent_around_ctx(get_log):

    @logger.indented("hey")
    @contextmanager
    def ctx():
        logger.info("001")
        yield
        logger.info("003")

    with ctx():
        logger.info("002")

    assert get_log() == "hey\n001\n002\n003\nDONE in no-time (hey)\n"



"""
This module is about Tokens. Tokens are string-based objects for when
``None`` is not enough, and ``Enum`` is too much.
Tokens are used to indicate a desired behavior, instead of some specific value.

Tokens can be created by simple instantiation::

    from easypy.tokens import Token
    AUTO = Token('AUTO')

If you don't like repeating yourself, however, you can use this 'dark' magic::

    from easypy.tokens import AUTO
"""

import sys
from types import ModuleType


__all__ = ["Token", "if_auto"]


class Token(str):
    """
    When ``None`` is not enough, and ``Enum`` is too much.
    Use to indicate a desired behavior, instead of some specific value:

        from easypy.tokens import AUTO, MAX

        def create_file(fname=AUTO, size=AUTO):

            if size is AUTO:
                size = get_capacity() / 2
            elif size is MAX:
                size = get_capacity()

            if fname is AUTO:
                from .random import random_nice_name
                fname = random_nice_name()
            os.truncate(fname, size)

    Also, to support use as cli options, a token can be compared with an str:

        AUTO == '<AUTO>'
        AUTO == 'AUTO'
        AUTO == '<auto>'
        AUTO == 'auto'

    """

    _all = {}

    def __new__(cls, name):
        name = name.strip("<>")
        try:
            return cls._all[name]
        except KeyError:
            pass
        cls._all[name] = self = super().__new__(cls, "<%s>" % name)
        return self

    def __repr__(self):
        return self

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self is other
        elif isinstance(other, str):
            # we allows this so that cli flags can be easily transformed into tokens (AUTO == 'auto')
            return self.strip("<>").lower() == other.strip("<>").lower()
        return False

    # we're already case insensitive when comparing
    def lower(self):
        return self

    def upper(self):
        return self

    def __hash__(self):
        return super().__hash__()


def if_auto(val, auto):
    """
    Convenience for the popular ``auto if val is AUTO else val``

    Example:

        config.foo_level = 100

        def foo(level=AUTO):
            level = if_auto(level, config.foo_level)
            return level

        assert foo() == 100
        assert foo(AUTO) == 100
        assert foo(1) == 1

    """
    AUTO = Token("AUTO")
    return auto if val is AUTO else val


class _TokensModule(ModuleType):
    """
    The module-hack that allows us to use ``from easypy.tokens import AUTO``
    """

    __all__ = ()  # to make help() happy
    __package__ = __name__
    _orig_module = sys.modules[__name__]

    def __getattr__(self, attr):
        try:
            return getattr(self._orig_module, attr)
        except AttributeError:
            pass

        if attr.startswith("_") or attr == 'trait_names':
            raise AttributeError(attr)

        token = Token("<%s>" % attr)
        setattr(self._orig_module, attr, token)
        return token

    def __dir__(self):
        return sorted(dir(self._orig_module) + list(Token._all))

    __path__ = []
    __file__ = __file__


mod = _TokensModule(__name__, __doc__)
sys.modules[__name__] = mod

del ModuleType
del _TokensModule
del mod, sys



from collections import defaultdict
from functools import wraps
from inspect import isgeneratorfunction, signature, Parameter
from contextlib import contextmanager as orig_contextmanager
from contextlib import ExitStack
from contextlib import _GeneratorContextManager


class KeyedStack(ExitStack):
    def __init__(self, context_factory):
        self.context_factory = context_factory
        self.contexts_dict = defaultdict(list)
        super().__init__()

    def enter_context(self, *key):
        cm = self.context_factory(*key)
        self.contexts_dict[key].append(cm)
        super().enter_context(cm)

    def exit_context(self, *key):
        self.contexts_dict[key].pop(-1).__exit__(None, None, None)


class _BetterGeneratorContextManager(_GeneratorContextManager):
    "This helper can handle generators and other context-managers"

    def __call__(self, func):
        if isgeneratorfunction(func):
            def inner(*args, **kwds):
                with self._recreate_cm():
                    yield from func(*args, **kwds)
        elif is_contextmanager(func):
            @contextmanager
            def inner(*args, **kwds):
                with self._recreate_cm():
                    with func(*args, **kwds) as ret:
                        yield ret
        else:
            def inner(*args, **kwds):
                with self._recreate_cm():
                    return func(*args, **kwds)
        return wraps(func)(inner)


# Some python version have a different signature for '_GeneratorContextManager.__init__', so we must adapt:
if signature(_GeneratorContextManager).parameters['args'].kind is Parameter.VAR_POSITIONAL:
    def contextmanager(func):
        @wraps(func)
        def helper(*args, **kwds):
            return _BetterGeneratorContextManager(func, *args, **kwds)
        return helper
else:
    def contextmanager(func):
        @wraps(func)
        def helper(*args, **kwds):
            return _BetterGeneratorContextManager(func, args, kwds)
        return helper

contextmanager.__doc__ = """@contextmanager decorator.

    Typical usage::

        @contextmanager
        def ctx(<arguments>):
            <setup>
            try:
                yield <value>
            finally:
                <cleanup>

    In a ``with`` statement::

        with ctx(<arguments>) as <variable>:
            <body>


    As a decorator for a function/method::

        @ctx(<arguments>)
        def simple_function():
            <do-something>

    As a decorator for a generator::

        @ctx(<arguments>)
        def generator():
            yield <something>

    As a decorator for a context manager (only those created using the @contextmanager decorator)::

        @ctx(<arguments>)
        @contextmanager
        def some_other_context_manager():
            <setup>
            try:
                yield <value>
            finally:
                <cleanup>
"""


# we use these to identify functions decorated by 'contextmanager'
_ctxm_code_samples = {
    f(None).__code__ for f in
    [contextmanager, orig_contextmanager]}


def is_contextmanager(func):
    return getattr(func, "__code__", None) in _ctxm_code_samples


@contextmanager
def breakable_section():
    """
    Useful for getting out of some deep nesting, as an alternative to a closure:

        item = None
        with breakable_section() as Break:
            if alpha:
                item = alpha.value
                raise Break

            if beta:
                for opt in beta.items:
                    if opt.is_the_one:
                        item = opt.value
                        raise Break

    Note that each 'Break' class this context-manager yield is unique,
    i.e it will only be caught by the context-manager that created it:

        with breakable_section() as Break1:

            with breakable_section() as Break2:
                raise Break1

            assert False  # will not reach here
    """
    Break = type("Break", (Exception,), {})
    try:
        yield Break
    except Break:
        pass



#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='real-easypy',
    version='0.4.2',
    description='easypy is a collection of python modules that makes developers happy',
    author='Ofer Koren',
    author_email='koreno@gmail.com',
    url='https://github.com/real-easypy/easypy',
    license='BSD',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'eziplog=easypy.ziplog:main',
            'ezcolorize=easypy.colors:main',
        ]
    },
)


# how to upload a package:
# 0. increment the version above
# 1. python3 setup.py sdist bdist_wheel
# 2. python3 -m twine upload dist/*



"""
This module is all about its 'aliases' (v.) class decorator
"""

from itertools import chain


# TODO - remove this, we don't need the python2 compatibility anymore
def super_dir(obj):
    """
    A python2/3 compatible way of getting the default ('super') behavior of __dir__
    """
    return sorted(set(chain(dir(type(obj)), obj.__dict__)))


# Python 3.4 does not have RecursionError - it throws a RuntimeError instead
try:
    _RecursionError = RecursionError
except NameError:
    _RecursionError = RuntimeError


class AliasingMixin():
    @property
    def _aliased(self):
        try:
            return getattr(self, self._ALIAS)
        except AttributeError:
            raise RuntimeError("object %r does no contain aliased object %r" % (self, self._ALIAS))

    def __dir__(self):
        members = set(super_dir(self))
        members.update(n for n in dir(self._aliased) if not n.startswith("_"))
        return sorted(members)

    def __getattr__(self, attr):
        if attr.startswith("_"):
            raise AttributeError(attr)
        try:
            return getattr(self._aliased, attr)
        except _RecursionError as e:
            if type(e) is RuntimeError and str(e) != 'maximum recursion depth exceeded':
                raise
            raise _RecursionError('Infinite recursion trying to access {attr!r} on {obj!r} (via {type_name}.{alias}.{attr})'.format(
                attr=attr,
                obj=self,
                type_name=type(self).__name__,
                alias=self._ALIAS))


def aliases(name, static=True):
    """
    A class decorator that makes objects of a class delegate to an object they contain.
    Inspired by D's "alias this".

    Example::

        class B():
            def foo(self):
                print('foo')

        @aliases('b')
        class A():
            b = B()

        a = A()
        a.foo()  # delegated to b.foo()


        @aliases('b', static=False)
        class A():
            def __init__(self):
                self.b = B()

        a = A()
        a.foo()  # delegated to b.foo()

    """
    def deco(cls):
        assert not static or hasattr(cls, name)
        return type(cls.__name__, (cls, AliasingMixin), dict(_ALIAS=name, __module__=cls.__module__))
    return deco



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import threading
import time
import itertools
import codecs
import re
import traceback
from math import ceil
from collections import namedtuple
from contextlib import contextmanager
from io import StringIO
from datetime import datetime, timedelta

from easypy.bunch import Bunch, bunchify
from easypy.colors import Colorizer
from easypy.collections import ilistify
from easypy.misc import clamp


def compact(line, length, ellipsis="....", suffix_length=20):
    if len(line) <= length:
        return line
    return line[:length-len(ellipsis)-suffix_length] + ellipsis + line[-suffix_length:]


is_non_printable = (set(range(32)) | set(range(127, 256))).__contains__

printables = set(map(ord, '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~ '))


def is_printable(char, include_space=True):
    """
    Returns ``True`` if the character is printable and NOT a control character.
    """
    if char == " " and not include_space:
        return False
    return char in printables


def bool_to_yesno(b):
    return "yes" if b else "no"


def yesno_to_bool(s):
    s = s.lower()
    if s not in ("yes", "no", "true", "false", "1", "0"):
        raise ValueError("Unrecognized boolean value: %r" % (s,))
    return s in ("yes", "true", "1")


def time_duration(delta, ago=False):
    if not isinstance(delta, timedelta):
        delta = timedelta(seconds=delta)

    negative = False
    if delta.total_seconds() < 0:
        negative = True
        delta = timedelta(seconds=-delta.total_seconds())

    days = delta.days
    hours, seconds = divmod(delta.seconds, 60*60)
    minutes, seconds = divmod(seconds, 60)

    if delta.total_seconds() <= 2:
        if ago:
            txt = "just now"
        else:
            txt = "no-time"
    else:
        if days:
            units = "a", "day", days, "hour", hours, 24
        elif hours:
            units = "an", "hour", hours, "minute", minutes, 60
        else:
            units = "a", "minute", minutes, "second", seconds, 60

        article, major_unit, major, minor_unit, minor, ratio = units
        distance = minor / ratio
        if distance > 0.6:
            major += 1
            minor -= ratio
            distance -= 1

        if not major:
            txt = "{minor} {minor_unit}s"
        elif major == 1:
            if distance <= -0.1:
                txt = "almost {article} {major_unit}"
            elif distance <= 0.1:
                txt = "{article} {major_unit}"
            elif distance <= 0.3:
                txt = "about {article} {major_unit}"
            elif distance <= 0.7:
                txt = "{major}½ {major_unit}s"
        else:
            if distance <= -0.1:
                txt = "almost {major} {major_unit}s"
            elif distance <= 0.1:
                txt = "{major} {major_unit}s"
            elif distance <= 0.3:
                txt = "about {major} {major_unit}s"
            elif distance <= 0.7:
                txt = "{major}½ {major_unit}s"
        txt = txt.format(**locals())
        if ago:
            txt += " from now" if negative else " ago"
    return txt


def from_utc(utc_str):
    return datetime.strptime(utc_str + "+0000", '%Y-%m-%dT%H:%M:%S.%fZ%z').timestamp()


def time_ago(t, now=None):
    now = datetime.fromtimestamp(now) if now else datetime.now()
    delta = now - datetime.fromtimestamp(t)
    return time_duration(delta, ago=True)


INPUT_TIME_PATTERNS = [
    ((re.compile(p[0]),) + tuple(p[1:])) for p in (
        (r"(\d{1,2}/\d{1,2}-\d{1,2}:\d{1,2}:\d{1,2})", "%d/%m-%H:%M:%S", 1, 0),        # 21/05-17:09:59
        (r"(\d{4}-\d{1,2}-\d{1,2}(?P<sep>[T-])\d{1,2}:\d{1,2}:\d{1,2})(?:\d*\.\d*)?(?P<utc>Z)?",
            "%Y-%m-%d{sep}%H:%M:%S", 0, time.timezone),
        (r"(\d{4}-\d{1,2}-\d{1,2}-\d{1,2}:\d{1,2})", "%Y-%m-%d-%H:%M", 0, 0),
        (r"(\d{1,2}-\d{1,2}-\d{1,2}:\d{1,2})", "%m-%d-%H:%M", 1, 0),
        (r"(\w+\s+\d{1,2}\s+\d{1,2}:\d{1,2}:\d{1,2})", "%b %d %H:%M:%S", 1, 0),
        (r"(\d{1,2}-\d{1,2}:\d{1,2}:\d{1,2})", "%d-%H:%M:%S", 2, 0),
        (r"(\d{1,2}-\d{1,2}:\d{1,2})", "%d-%H:%M", 2, 0),
        (r"(\d{1,2}:\d{1,2}:\d{1,2})", "%H:%M:%S", 3, 0),
        (r"(\d{1,2}:\d{1,2})", "%H:%M", 3, 0),
    )]


def parse_fuzzy_time(ts, baseline=None):
    for regex, fmt, missing, offset in INPUT_TIME_PATTERNS:
        match = regex.match(ts)
        if match:
            break
    else:
        raise ValueError("Invalid time pattern: %r (must be one of: [%s])" % (ts, ", ".join(a[1] for a in INPUT_TIME_PATTERNS)))

    ts = match.group(1)
    params = match.groupdict()
    if not params.get('utc'):
        offset = 0
    time_tuple = time.strptime(ts, fmt.format(**params))
    if missing:
        baseline = time.localtime(baseline)
        time_tuple = baseline[:missing] + time_tuple[missing:]
    return time.mktime(time_tuple) - offset


Node = namedtuple("Node", "fmt args kwargs children")


class IndentableTextBuffer():
    NICE_BOX = bunchify(
        LINE=            "─",
        INDENT_SEGMENT=  "│   ",
        OPEN=            "┬",
        INDENT_OPEN=     "├───",
        INDENT_CLOSE=    "╰",
        SEGMENT_END=     "╼ ",
        SEGMENT_START=   " ╾",
        SECTION_OPEN=    "╮",
        SECTION_SEGMENT= "│",
        SECTION_CLOSE=   "╯",
        )
    TEXTUAL_BOX = bunchify(
        LINE=            "-",
        INDENT_SEGMENT=  "|   ",
        INDENT_OPEN=     "+---",
        OPEN=            ".",
        INDENT_CLOSE=    "`",
        SEGMENT_END=     "- ",
        SEGMENT_START=   " -",
        SECTION_OPEN=    ".",
        SECTION_SEGMENT= "|",
        SECTION_CLOSE=   "*",
        )

    def __init__(self, fmt="", *args, **kwargs):
        self.current = self.root = Node(fmt, args, kwargs, [])

    def __repr__(self):
        return self.render(width=80)

    def __len__(self):
        def count(root):
            for child in root.children:
                if isinstance(child, str):
                    yield 1
                else:
                    yield from count(child)
        return sum(count(self.root))

    def write(self, line, *args, **kwargs):
        if args or kwargs:
            line = line.format(*args, **kwargs)
        self.current.children.append(line)

    def extend(self, other):
        self.current.children.extend(other.root.children)

    @contextmanager
    def indent(self, fmt, *args, **kwargs):
        parent = self.current
        self.current = Node(fmt, args, kwargs, [])
        parent.children.append(self.current)
        yield
        self.current = parent

    def render(self, width=None, textual=None, prune=False, file=None, overflow='ignore', edges=True):
        if width is None:
            from .logging import TERM_WIDTH as width
        if textual is None:
            from .logging import GRAPHICAL
            textual = not GRAPHICAL

        buff = file if file else StringIO()
        G = self.TEXTUAL_BOX if textual else self.NICE_BOX
        from textwrap import wrap

        def has_descendents(elem):
            if isinstance(elem, str):
                return True
            return any(map(has_descendents, elem.children))

        def write_tree(elem, depth=0):
            if isinstance(elem, str):
                prefix = G.INDENT_SEGMENT * (depth-1)
                prefix += G.INDENT_SEGMENT
                for par in elem.splitlines():
                    if overflow == "wrap":
                        lines = wrap(par, width-len(prefix)-1)
                    elif overflow == "trim":
                        lines = [compact(par, width-len(prefix)-1)]
                    else:
                        lines = [par]
                    for line in lines:
                        line = prefix + line
                        if len(line) < width:
                            if edges:
                                line = line.ljust(width-1) + G.SECTION_SEGMENT
                        buff.write(line + "\n")
            elif not prune or has_descendents(elem):
                if depth:
                    prefix = G.INDENT_SEGMENT * (depth-1) + G.INDENT_OPEN
                else:
                    prefix = ""
                txt = (G.SEGMENT_END + elem.fmt.format(*elem.args, **elem.kwargs) + G.SEGMENT_START) if elem.fmt else ""
                header = prefix + G.OPEN + txt
                buff.write(header.ljust(width-1, G.LINE) + G.SECTION_OPEN + "\n")
                for child in elem.children:
                    write_tree(child, depth+1)
                footer = (G.INDENT_SEGMENT*depth + G.INDENT_CLOSE + G.LINE * (width-len(G.INDENT_SEGMENT)*(depth+1)-len(txt)+2) + txt)
                buff.write(footer + G.SECTION_CLOSE + "\n")

        write_tree(self.root)
        if not file:
            return buff.getvalue()


def format_in_columns(elements, total_width=None, sep="  ", indent="  ", min_height=10):
    """
    >>> print(format_in_columns([str(i) for i in range(100)], 50))
      0  10  20  30  40  50  60  70  80  90
      1  11  21  31  41  51  61  71  81  91
      2  12  22  32  42  52  62  72  82  92
      3  13  23  33  43  53  63  73  83  93
      4  14  24  34  44  54  64  74  84  94
      5  15  25  35  45  55  65  75  85  95
      6  16  26  36  46  56  66  76  86  96
      7  17  27  37  47  57  67  77  87  97
      8  18  28  38  48  58  68  78  88  98
      9  19  29  39  49  59  69  79  89  99
    """

    if not total_width:
        try:
            total_width, _ = os.get_terminal_size()
        except:
            total_width = 80

    widest = min(max(len(k) for k in elements), total_width)
    columns = max((total_width - len(indent)) // (widest + len(sep)), 1)
    height = max(min_height, (len(elements) // columns) + 1)

    # arrange the elements in columns
    columns = [[elem for (__, elem) in group]
               for __, group in itertools.groupby(enumerate(elements), lambda p: p[0]//height)]
    rows = itertools.zip_longest(*columns)

    col_max = total_width - len(sep) * (len(columns) - 1)
    column_lens = [min(max(map(len, column)), col_max) for column in columns]

    return '\n'.join(indent + sep.join([(string or "").ljust(column_lens[column_num])
                                        for column_num, string in enumerate(row)])
                     for row in rows)


def format_dict(obj, max_width=200, indent="  "):
    key_width = val_width = 0
    items = []
    key_fmt, sep, format_output = ((str, " = ", "\n{0.__class__.__name__}(\n{1}\n)".format)
                                   if isinstance(obj, Bunch) else
                                   (repr, " : ", "\n{{\n{1}\n}}".format))
    for key, val in obj.items():
        key = key_fmt(key)
        val = repr(val)
        items.append((key, val))
        # figure out widest key/val
        key_width = max(key_width, len(key))
        val_width = max(val_width, len(val), 5)

    # truncate to terminal width
    key_width = min(key_width, max_width - len(sep))
    val_width = min(val_width, max_width - len(sep) - key_width - len(indent))
    elements = ["{:{key_width}}{}{:{val_width}},".format(k, sep, v, key_width=key_width, val_width=val_width)
                for k, v in sorted(items)]
    return format_output(obj, format_in_columns(elements, max_width, indent=indent))


# HexDump-Related ######################################################################

to_hex = {i: ('%02x' % i)
          for i in range(256)
          }.__getitem__

to_printable = {ch: chr(ch) if is_printable(ch) else '.'
                for ch in range(256)
                }.__getitem__


def format_hex(buff, chunk_size=8):
    return '  '.join(
        ' '.join(to_hex(char)
                 for char in buff[offset : (offset + chunk_size) if chunk_size else None])
        for offset in range(0, len(buff), chunk_size or len(buff))
    )


def format_printable(buff, chunk_size=8):
    return ' '.join(
        ''.join(to_printable(char)
                for char in buff[offset : (offset + chunk_size) if chunk_size else None])
        for offset in range(0, len(buff), chunk_size or len(buff))
    )


def iter_hexdump(data, bytes_per_line=32, chunk_size=4, skip_repeats=True, start_offset=0):
    rows = []
    skipping = False
    previous_content = None
    rows.append(("0x....+",
                 format_hex(range(bytes_per_line), chunk_size=chunk_size).upper().replace(" 0", "  "),
                 format_printable([ord("-")]*bytes_per_line, chunk_size=chunk_size)))
    for offset in range(0, len(data), bytes_per_line):
        content = data[offset:offset+bytes_per_line]
        if skip_repeats and (0 < offset < (len(data)-bytes_per_line)) and content == previous_content:
            if skipping:
                continue
            skipping = True
            row = '*'.ljust(7), "", ""
        else:
            previous_content = content
            skipping = False
            row = ("0x{:04x}:".format(start_offset+offset),
                   format_hex(content, chunk_size),
                   format_printable(content, chunk_size)
                   )
        rows.append(row)
    hex_width = len(rows[0][1]) if rows else 0
    fmt = "{} {:{hex_width}} | {}".format
    return (fmt(*row, hex_width=hex_width) for row in rows)


def hexdump(data, bytes_per_line=32, chunk_size=4, skip_repeats=True, start_offset=0):
    return ('\n'.join(iter_hexdump(data, bytes_per_line, chunk_size, skip_repeats, start_offset)))


def from_hexdump(data):
    from io import StringIO, BytesIO
    buff = BytesIO()
    last_offset = None
    repeats = False
    to_bytes = bytes
    for line in StringIO(data):
        if not line:
            continue
        if line.startswith("0x....+"):
            continue
        parts = line.split()
        offset = parts[0]
        if offset.startswith("*"):
            repeats = True
        else:
            offset = eval(offset[:-1])
            if repeats:
                buff.write(chunks * ((offset-last_offset)//len(chunks)-1))
                repeats = False
            chunks = to_bytes(int(c, 16) for c in itertools.takewhile(lambda c: c != "|", parts[1:]))
            buff.write(chunks)
            last_offset = offset
    return buff.getvalue()


# Stateless encoder/decoder
class HexDumpCodec(codecs.Codec):
    def __init__(self, *args):
        self.args = args

    def decode(self, data, errors='strict'):
        return hexdump(data, *self.args), len(data)

    def encode(self, data, errors='strict'):
        return from_hexdump(data), len(data)


# Register the codec search function
def find_hexdump(encoding):
    """Return the codec for ``HexDump``.
    """
    if encoding.startswith('hexdump'):
        args = list(map(int, encoding.split("_")[1:]))
        return codecs.CodecInfo(
            name='hexdump',
            encode=HexDumpCodec(*args).encode,
            decode=HexDumpCodec(*args).decode,
            )
    return None

codecs.register(find_hexdump)


def to_new_style_formatter(string):
    def repl(matched):
        keyword = matched.group(1)
        if keyword:
            return "{%s}" % keyword.strip('()')
        else:
            return "{%d}" % next(index)
    index = itertools.count()
    return re.sub(r'(?:%(\(\w+\))?[diouxXeEfFgGcrs])', repl, string)


class TrimmingTemplate(str):
    """
    Use new-style string formatting syntax with the old modulu string formatting operator:

    >>> TrimmingTemplate("{b}: {a}") % dict(a=1, b=2)
    '2: 1'


    Use the ``~`` next to the field width specifier so that the content is
    trimmed to fit the specified field width.  Works with string data only.

    >>> TrimmingTemplate("{id:5}:{header:10~} {footer:~11}") % dict(id=225, header='This is not the end', footer='This is not the end!')
    '  225:This is... ...the end!'

    """

    RE_NEW_STYLE_FORMATTERS = re.compile(r"{(\w+):([^}]*)}")

    def __init__(self, s):
        assert s == to_new_style_formatter(s), "Can't use old-style string formatting syntax"
        self.trimmers = {}
        for name, format in self.RE_NEW_STYLE_FORMATTERS.findall(s):
            if format.startswith("~"):
                limiter = lambda s, limit=int(format[1:]): ("..." + s[-(limit-3):]) if len(s) > limit else s
            elif format.endswith("~"):
                limiter = lambda s, limit=int(format[:-1]): s[:limit-3] + "..." if len(s) > limit else s
            else:
                continue
            self.trimmers[name] = limiter
        self._fmt = self.replace("~", "")

    def __mod__(self, other):
        if isinstance(other, dict):
            return self.format(**other)
        elif isinstance(other, tuple):
            return self.format(*other)
        else:
            return self.format(other)

    def format(self, **kwargs):
        kwargs = {k: self.trimmers.get(k, lambda s: s)(v) for (k, v) in kwargs.items()}
        return self._fmt.format(**kwargs)


# Tracebacks
def get_traceback_formatter(default_root_path=None):
    from plumbum import local
    if default_root_path:
        def relative(path):
            r = local.path(path) - default_root_path
            if r and r[0] != "..":
                path = "./%s" % r
            return path
        default_root_path = local.path(default_root_path)
    else:
        relative = lambda path: path

    def _format_list_iter(extracted_list):
        extracted_list = list(extracted_list)
        lines = []
        for filename, lineno, name, line in extracted_list:
            filename = relative(filename)
            left = "  {}:{} ".format(filename, lineno)
            right = " {}".format(name)
            lines.append((len(left)+len(right), left, right, line))

        width = max(args[0] for args in lines) + 4
        for _, left, right, line in lines:
            item = left.ljust(width-len(right), ".") + right
            if line:
                item = item + ' >> {}'.format(line.strip())
            yield item + '\n'
    return _format_list_iter


def format_tb(*args, **kw):
    import traceback
    default_root_path = kw.pop("default_root_path", None)
    orig, traceback._format_list_iter = traceback._format_list_iter, get_traceback_formatter(default_root_path=default_root_path)
    try:
        return traceback.format_tb(*args)
    finally:
        traceback._format_list_iter = orig


def format_thread_stack(frame, skip_modules=[threading]):
    stack = traceback.extract_stack(frame)
    if skip_modules:
        itr_stack = iter(stack)
        items = []
        fnames = {m.__file__ for m in ilistify(skip_modules)}
        # skip everything until after specified module
        for fname, *_ in itr_stack:
            if fname in fnames:
                items.append([fname] + _)
                for i, (fname, *_) in enumerate(itr_stack, 1):
                    if fname not in fnames:
                        if i > 1:
                            items.append([last_fname, "...(%s)" % i, "---", "---"])
                        break
                    last_fname = fname
            items.append([fname] + _)
        if len(items) <= 2:
            items = stack
    else:
        items = stack

    return ''.join(traceback.format_list(items))


def format_table(table, titles=True):
    """
    Returns a multilined string representing the given table (2d list) as a table, with equal-width columns.

    :param titles: if true, the first row in the table is taken as headers for
    the table, adding a separator on the second line
    """

    fmt_simple = "{:{width}}"
    fmt_string = "{!s:{width}}"  # for type that don't accept a ':width' specifier when formatted (NoneType, dict, ...)

    def safe_format(cell, width=1):
        if isinstance(cell, bool):
            # because if we force it to str it becomes an int (True -> 1, False -> 0)
            return fmt_string.format(cell, width=width)

        try:
            return fmt_simple.format(cell, width=width)
        except TypeError:
            return fmt_string.format(cell, width=width)

    widths = [max(map(len, map(safe_format, column))) for column in zip(*table)]
    txt = ''
    for i, row in enumerate(table):
        if titles and i == 1:
            txt += '-'.join('-' * width for width in widths) + '\n'
        txt += '|'.join(safe_format(cell, width=width) for cell, width in zip(row, widths))
        txt += '\n'
    return txt


def format_size(num, suffix='B'):
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def ipython_fields_repr(name, field_items, p, cycle):
    from easypy.colors import DARK_CYAN
    if cycle:
        p.text('%s(...)' % (name,))
        return
    prefix = '%s(' % (name,)
    with p.group(len(prefix), prefix, ')'):
        for idx, (k, v) in enumerate(field_items):
            if idx:
                p.text(',')
                p.breakable()
            with p.group(len(k)+1, DARK_CYAN(k) + "="):
                p.pretty(v)


def ipython_mapping_repr(mapping, p, cycle):
    """Used by IPython. add to any mapping class as '_repr_pretty_'"""
    ipython_fields_repr(mapping.__class__.__name__, sorted(mapping.items()), p, cycle)


BAR_SEQUENCE = ' ▏▎▍▌▋▊▉██'
VERT_SEQUENCE = ' ▁▂▃▄▅▆▇█'
LITTLE_DIGITS = '⁰¹²³⁴⁵⁶⁷⁸⁹'
RULER_MARKS = '0' + LITTLE_DIGITS[1:]


def vertbar(ratio):
    ratio = max(min(1.0, ratio), 0.0)
    return VERT_SEQUENCE[ceil(clamp(ratio, 0, 1) * (len(VERT_SEQUENCE)-1))]


def horizbar(ratio, width=2):
    ratio = max(min(1.0, ratio), 0.0) * width
    return "".join(
        BAR_SEQUENCE[
            ceil(clamp(ratio-p, 0, 1) * (len(BAR_SEQUENCE)-1))
            ]
        for p in range(width))


def name_generator():
    from string import digits
    from itertools import product

    easynames = ('alfa bravo charlie delta echo foxtrot golf hotel india juliett kilo lima mike november '
                 'oscar papa quebec romeo sierra tango uniform victor whiskey xray yankee zulu').split()

    for parts in product(easynames, easynames, digits):
        yield "_".join(parts)


def percentages_comparison(actual, expected, key_caption='Item', color_bounds={'green': 0,
                                                                               'yellow': 5,
                                                                               'red': 10}):
    from easypy.tables import Table, Column
    assert all(0 <= count for count in actual.values())
    assert all(0 <= count for count in expected.values())

    table = Table()
    table.add_column(Column("key", title=key_caption))
    table.add_column(Column("actual", title='Actual'))
    table.add_column(Column("expected", title='Expected'))
    table.add_column(Column("diff", title='Difference'))

    actual_total = sum(actual.values())
    expected_total = sum(expected.values())
    if actual_total == 0 or expected_total == 0:
        return table

    color_bounds = sorted(color_bounds.items(), key=lambda a: a[1], reverse=True)

    def color_for(percentage):
        for color, bound in color_bounds:
            if bound <= percentage:
                return color

    def generate():
        for key in actual.keys() | expected.keys():
            actual_count = actual.get(key, 0)
            expected_count = expected.get(key, 0)
            if actual_count == 0 and expected_count == 0:
                continue
            actual_percentage = 100 * actual_count / actual_total
            expected_percentage = 100 * expected_count / expected_total
            diff = actual_percentage - expected_percentage
            abs_diff = abs(diff)
            color = color_for(abs_diff)
            if color:
                colorizer = Colorizer(color)
                yield Bunch(key=colorizer(key),
                            actual=colorizer('%.1f%%' % actual_percentage),
                            expected=colorizer('%.1f%%' % expected_percentage),
                            diff=colorizer('%.1f%%' % diff),
                            abs_diff=abs_diff)

    for item in sorted(generate(), key=lambda a: a.abs_diff, reverse=True):
        table.add_row(**item)
    return table


class _ReprAsString:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return str(self.value)
